#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for the schooltool.timetable.term module.
"""
import calendar
import unittest
import doctest
from pytz import timezone
from datetime import date, datetime

from zope.component import provideAdapter
from zope.interface import Interface
from zope.interface import implements
from zope.interface.verify import verifyObject
from zope.container.contained import Contained
from zope.location.interfaces import ILocation
from zope.app.testing.setup import placefulSetUp, placefulTearDown
from zope.app.testing.setup import placelessSetUp, placelessTearDown

from schooltool.schoolyear.interfaces import ISchoolYearContainer
from schooltool.schoolyear.schoolyear import SchoolYear
from schooltool.schoolyear.schoolyear import getSchoolYearContainer
from schooltool.term.term import getTermContainer
from schooltool.term.interfaces import ITermContainer
from schooltool.term import interfaces, term
from schooltool.testing import setup


class TermStub(Contained):

    implements(interfaces.ITerm)

    #     November 2003
    #  Su Mo Tu We Th Fr Sa
    #                     1
    #   2  3  4  5  6  7  8
    #   9 10 11 12 13 14 15
    #  16 17 18 19<20-21>22
    #  23<24-25-26>27 28 29
    #  30

    first = date(2003, 11, 20)
    last = date(2003, 11, 26)

    def __iter__(self):
        return iter((date(2003, 11, 20),
                     date(2003, 11, 21),
                     date(2003, 11, 22),
                     date(2003, 11, 23),
                     date(2003, 11, 24),
                     date(2003, 11, 25),
                     date(2003, 11, 26)))

    def isSchoolday(self, day):
        return day in (date(2003, 11, 20),
                       date(2003, 11, 21),
                       date(2003, 11, 24),
                       date(2003, 11, 25),
                       date(2003, 11, 26))

    def __contains__(self, day):
        return date(2003, 11, 20) <= day <= date(2003, 11, 26)


class TestTerm(unittest.TestCase):

    def test_interface(self):

        cal = term.Term('Sample', date(2003, 9, 1), date(2003, 12, 24))
        verifyObject(interfaces.ITerm, cal)
        verifyObject(interfaces.ITermWrite, cal)
        verifyObject(ILocation, cal)

    def testAddRemoveSchoolday(self):
        cal = term.Term('Sample', date(2003, 9, 1), date(2003, 9, 14))

        self.assert_(not cal.isSchoolday(date(2003, 9, 1)))
        self.assert_(not cal.isSchoolday(date(2003, 9, 2)))
        self.assertRaises(ValueError, cal.isSchoolday, date(2003, 9, 15))

        cal.add(date(2003, 9, 2))
        self.assert_(cal.isSchoolday(date(2003, 9, 2)))
        cal.remove(date(2003, 9, 2))
        self.assert_(not cal.isSchoolday(date(2003, 9, 2)))
        self.assertRaises(ValueError, cal.add, date(2003, 9, 15))
        self.assertRaises(ValueError, cal.remove, date(2003, 9, 15))

    def testReset(self):
        cal = term.Term('Sample', date(2003, 9, 1), date(2003, 9, 15))
        cal.addWeekdays(1, 3, 5)

        new_first, new_last = date(2003, 8, 1), date(2003, 9, 30)
        cal.reset(new_first, new_last)

        self.assertEqual(cal.first, new_first)
        self.assertEqual(cal.last, new_last)
        for d in cal:
            self.assert_(not cal.isSchoolday(d))

        self.assertRaises(ValueError, cal.reset, new_last, new_first)

    def testMarkWeekday(self):
        cal = term.Term('Sample', date(2003, 9, 1), date(2003, 9, 17))
        for day in 1, 8, 15:
            self.assert_(not cal.isSchoolday(date(2003, 9, day)))

        cal.addWeekdays(calendar.MONDAY)
        for day in 1, 8, 15:
            self.assert_(cal.isSchoolday(date(2003, 9, day)))
            self.assert_(not cal.isSchoolday(date(2003, 9, day+1)))

        cal.removeWeekdays(calendar.MONDAY, calendar.TUESDAY)
        for day in 1, 8, 15:
            self.assert_(not cal.isSchoolday(date(2003, 9, day)))
            self.assert_(not cal.isSchoolday(date(2003, 9, day+1)))

        cal.addWeekdays(calendar.MONDAY, calendar.TUESDAY)
        for day in 1, 8, 15:
            self.assert_(cal.isSchoolday(date(2003, 9, day)))
            self.assert_(cal.isSchoolday(date(2003, 9, day+1)))

        cal.toggleWeekdays(calendar.TUESDAY, calendar.WEDNESDAY)
        for day in 1, 8, 15:
            self.assert_(cal.isSchoolday(date(2003, 9, day)))
            self.assert_(not cal.isSchoolday(date(2003, 9, day+1)))
            self.assert_(cal.isSchoolday(date(2003, 9, day+2)))

    def test_contains(self):
        cal = term.Term('Sample', date(2003, 9, 1), date(2003, 9, 16))
        self.assert_(date(2003, 8, 31) not in cal)
        self.assert_(date(2003, 9, 17) not in cal)
        for day in range(1, 17):
            self.assert_(date(2003, 9, day) in cal)
        self.assertRaises(TypeError, cal.__contains__, 'some string')


class TestGetTermForDate(unittest.TestCase):

    def setUp(self):
        placefulSetUp()
        provideAdapter(getTermContainer, [Interface], ITermContainer)
        provideAdapter(getSchoolYearContainer)
        app = setup.setUpSchoolToolSite()

        schoolyear = SchoolYear("Sample", date(2004, 9, 1), date(2005, 12, 20))
        ISchoolYearContainer(app)['2004-2005'] = schoolyear

        self.term1 = term.Term('Sample', date(2004, 9, 1), date(2004, 12, 20))
        self.term2 = term.Term('Sample', date(2005, 1, 1), date(2005, 6, 1))
        terms = ITermContainer(app)
        terms['2004-fall'] = self.term1
        terms['2005-spring'] = self.term2
        self.app = app

    def tearDown(self):
        placefulTearDown()

    def test_getTermForDate(self):
        self.assert_(term.getTermForDate(date(2004, 8, 31)) is None)
        self.assert_(term.getTermForDate(date(2004, 9, 1)) is self.term1)
        self.assert_(term.getTermForDate(date(2004, 11, 5)) is self.term1)
        self.assert_(term.getTermForDate(date(2004, 12, 20)) is self.term1)
        self.assert_(term.getTermForDate(date(2004, 12, 21)) is None)
        self.assert_(term.getTermForDate(date(2005, 3, 17)) is self.term2)
        self.assert_(term.getTermForDate(date(2005, 11, 5)) is None)

    def test_getNextTermForDate(self):
        self.assert_(term.getNextTermForDate(date(2004, 8, 31)) is self.term1)
        self.assert_(term.getNextTermForDate(date(2004, 9, 1)) is self.term1)
        self.assert_(term.getNextTermForDate(date(2004, 11, 5)) is self.term1)
        self.assert_(term.getNextTermForDate(date(2004, 12, 20)) is self.term1)
        self.assert_(term.getNextTermForDate(date(2004, 12, 21)) is self.term2)
        self.assert_(term.getNextTermForDate(date(2005, 3, 17)) is self.term2)
        self.assert_(term.getNextTermForDate(date(2005, 11, 5)) is self.term2)
        self.term3 = term.Term('Sample', date(2005, 9, 1), date(2005, 12, 20))
        terms = ITermContainer(self.app)
        terms['2005-fall'] = self.term3
        self.assert_(term.getNextTermForDate(date(2005, 8, 30)) is self.term3)
        self.assert_(term.getNextTermForDate(date(2004, 8, 31)) is self.term1)
        self.assert_(term.getNextTermForDate(date(2004, 12, 22)) is self.term2)
        del terms['2004-fall']
        del terms['2005-spring']
        del terms['2005-fall']
        self.assert_(term.getNextTermForDate(date(2004, 8, 31)) is None)


def doctest_DateManagerUtility_today():
    """Test for today.

    Today returns the date of today, according to the application
    prefered timezone:

        >>> from schooltool.term.term import DateManagerUtility
        >>> from schooltool.app.interfaces import IApplicationPreferences
        >>> dm = DateManagerUtility()
        >>> tz_name = "Europe/Vilnius"
        >>> class PrefStub(object):
        ...     @property
        ...     def timezone(self):
        ...         return tz_name

        >>> class STAppStub(dict):
        ...     def __init__(self, context):
        ...         pass
        ...     def __conform__(self, iface):
        ...         if iface == IApplicationPreferences:
        ...             return PrefStub()

        >>> from schooltool.app.interfaces import ISchoolToolApplication
        >>> provideAdapter(STAppStub, adapts=[None], provides=ISchoolToolApplication)

        >>> current_time = timezone('UTC').localize(datetime.utcnow())

        >>> tz_name = 'Pacific/Midway'
        >>> tz = timezone(tz_name)
        >>> today_date = current_time.astimezone(tz).date()
        >>> dm.today == today_date
        True

        >>> tz_name = 'Pacific/Funafuti'
        >>> tz = timezone('Pacific/Funafuti')
        >>> today_date = current_time.astimezone(tz).date()
        >>> dm.today == today_date
        True

    """


def createSchoolYearWith3Terms():
    sy = SchoolYear("Sample", date(2009, 9, 1), date(2010, 05, 20))
    sy['a-term1'] = term.Term('Term 1',
        date(2009, 9, 1), date(2009, 12, 20))
    sy['b-term3'] = term.Term('Term 3',
        date(2010, 3, 16), date(2010, 5, 20))
    sy['c-term2'] = term.Term('Term 2',
        date(2010, 1, 1), date(2010, 3, 15))
    return sy


def doctest_listTerms():
    """Test for listTerms.

    Stub adapter that obtains the term container.

        >>> sy = createSchoolYearWith3Terms()
        >>> provideAdapter(lambda context: sy, [Interface], ITermContainer)

    Show that listTerm lists terms chronologically.

        >>> context = object()
        >>> [t.title for t in term.listTerms(context)]
        ['Term 1', 'Term 2', 'Term 3']

    """


def doctest_getPreviousTerm_getNextTerm():
    """Test for getPreviousTerm and getNextTerm

    Stub adapter that obtains the term container.

        >>> sy = createSchoolYearWith3Terms()
        >>> provideAdapter(lambda context: sy, [Interface], ITermContainer)

        >>> def printTerm(term):
        ...     if term is None:
        ...         print 'None'
        ...     else:
        ...         print '%s:' % term.title, term.first, '-', term.last

        >>> app = object()
        >>> t1, t2, t3 = term.listTerms(app)

    So, we'll start with this term:

        >>> printTerm(t1)
        Term 1: 2009-09-01 - 2009-12-20

    And print several next terms.

        >>> printTerm(term.getNextTerm(t1))
        Term 2: 2010-01-01 - 2010-03-15

        >>> printTerm(term.getNextTerm(t2))
        Term 3: 2010-03-16 - 2010-05-20

    There are no terms after Term 3.

        >>> printTerm(term.getNextTerm(t3))
        None

    Now let's check previous terms.

        >>> printTerm(term.getPreviousTerm(t3))
        Term 2: 2010-01-01 - 2010-03-15

        >>> printTerm(term.getPreviousTerm(t2))
        Term 1: 2009-09-01 - 2009-12-20

    No terms before Term 1.

        >>> printTerm(term.getPreviousTerm(t1))
        None

    """


def setUp(test):
    placelessSetUp()


def tearDown(test):
    placelessTearDown()


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(doctest.DocTestSuite(setUp=setUp,
                                       tearDown=tearDown))
    suite.addTest(unittest.makeSuite(TestTerm))
    suite.addTest(unittest.makeSuite(TestGetTermForDate))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2006 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Interfaces for SchoolTool security policy.

"""

from zope.interface import Interface
from zope import schema
from zope.interface import Attribute
from zope.configuration.fields import PythonIdentifier
from zope.location.interfaces import IContained

from schooltool.common import SchoolToolMessage as _


class ICrowdsUtility(Interface):
    """Crowds Utility holds registered security information"""

    factories = schema.Dict(
        title=u"Crowd Factories",
        description=u"Maps crowd names to crowd factories")

    crowds = schema.Dict(
        title=u"Permission Crowds",
        description=u"Maps (permission, interface)s to crowd names")

    # TODO: update interface, it's out of date


class IDescriptionUtility(Interface):
    """Access rights description utility"""

    # TODO: describe fields


class ICrowd(Interface):
    """A crowd is conceptually a set of principals.

    A crowd need only support one operation -- a membership test.
    """

    title = schema.TextLine(title=u"Title", required=False)

    description = schema.Text(title=u"Description", required=False)

    def contains(principal):
        """Return True if principal is in the crowd."""


class IAccessControlCustomisations(Interface):
    """Access Control Customisation storage."""

    def get(key):
        """Return a value of a setting stored under the key."""

    def set(key, value):
        """Set the value of a setting stored under the key."""

    def __iter__():
        """Iterate through all customisation settings."""


class IAccessControlSetting(Interface):
    """An access control customisation setting."""

    key = PythonIdentifier(description=u"""A key that identified the setting.
                           For example: 'members_manage_groups',
                           'teachers_edit_person_info'
                           """)
    default = schema.Bool(title=u"The default value for the setting.")
    text = schema.TextLine(
        title=u"Description of the setting for the user interface.")
    alt_text = schema.TextLine(
        title=u"Description of the effect when the setting is off.")

    def getValue():
        """Return the value of the setting.

        Return the default if it is not set in the
        AccessControlCusomisations storage.
        """

    def setValue(value):
        """Set the value of the setting."""


class IDescription(IContained):
    """Base interface for a description object."""

    title = schema.TextLine(
        title=_("Title"),
        description=_("Title of the group."))

    description = schema.Text(
        title=_("Description"),
        required=False)


class IDescriptionGroup(IDescription):
    """A group of access descriptions."""


class IGroupAction(IDescription):
    """Actions that given permission enables on an object implementing
    given interface.
    """

    interface = Attribute(u"Interface")

    permission = schema.TextLine(
        title=u"Permission",
        required=True)


class ICrowdToDescribe(Interface):
    """Adapt to this interface to get the crowd to describe."""


class ICrowdDescription(IDescription):
    """Description of a crowd."""



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
"""
import unittest
import doctest
from pprint import pprint

from schooltool.app.browser.testing import setUp, tearDown
from schooltool.table.pdf import GridCell, Grid
from schooltool.table.pdf import makeCoordinateBlocks


def print_blocks(blocks):
    sx = min([b[0][0] for b in blocks])
    ex = max([b[1][0] for b in blocks])
    sy = min([b[0][1] for b in blocks])
    ey = max([b[1][1] for b in blocks])
    bmap = {}
    for n, block in enumerate(blocks):
        for x in range(block[0][0], block[1][0]+1):
            for y in range(block[0][1], block[1][1]+1):
                bmap[(x, y)] = n
    result = ''
    template = '%%%ds' % (len(str(n))+1)
    for y in range(sy, ey+1):
        result += '\n:'
        for x in range(sx, ex+1):
            result += template % bmap.get((x, y), '')
    print result


def test_makeCoordinateBlocks():
    """

    makeCoordinateBlocks takes coordinates and returns spanning blocks.

        >>> blocks = makeCoordinateBlocks([
        ...     (0,1), (1,1), (2,1), (3,1),
        ...     (0, 2), (2, 2),
        ...     (0, 3), (1, 3)
        ... ])

        >>> blocks
        [((0, 1), (3, 1)),
         ((0, 2), (0, 2)),
         ((2, 2), (2, 2)),
         ((0, 3), (1, 3))]

    We can see it prefers building horizontal blocks over vertical ones.

        >>> print_blocks(blocks)
        : 0 0 0 0
        : 1   2
        : 3 3

        >>> blocks = makeCoordinateBlocks([
        ... (0,1), (1,1), (2,1), (3,1),
        ... (0, 2), (1,2), (2, 2),
        ... (0, 3), (1,3),
        ... ])

        >>> print_blocks(blocks)
        : 0 0 0 0
        : 1 1 1
        : 2 2

    Blocks are created when it finds a line of coordinates that can be
    extruded "down".

        >>> print_blocks(makeCoordinateBlocks([
        ...     (1, 1), (1, 2), (2, 1), (2,2), (2, 3)]))
        : 0 0
        : 0 0
        :   1

        >>> print_blocks(makeCoordinateBlocks([
        ...     (1, 0), (1, 1), (1, 2), (2, 1), (2, 2), (4, 2), (2, 3),
        ...     (3, 1), (3, 2)]))
        : 0
        : 1 1 1
        : 1 1 1 2
        :   3

    Blocks are however made only when "upper" scanline can be extruded "down"

        >>> print_blocks(makeCoordinateBlocks([
        ...     (1, 0), (1, 1), (1, 2), (2, 1), (2, 2), (4, 1), (2, 3),
        ...     (3, 1), (3, 2)]))
        : 0
        : 1 1 1 1
        : 2 2 2
        :   3

    """


def doctest_Grid_getStyleDirectives():
    """

        >>> grid = Grid([], [],  {}, 0)

        >>> cell = GridCell('Hello',
        ...     font_name='Font',
        ...     font_size=816,
        ...     font_leading=7,
        ...     color='very_black',
        ...     background_color='very_white',
        ...     align='left',
        ...     valign='bottom',
        ...     )

        >>> directives = grid.getStyleDirectives(cell, 5, 10)

        >>> pprint(directives)
        {'blockAlignment': [('value', 'left')],
         'blockBackground': [('colorName', 'very_white')],
         'blockFont': [('leading', 7), ('name', 'Font'), ('size', 816)],
         'blockTextColor': [('colorName', 'very_black')],
         'blockValign': [('value', 'bottom')]}

        >>> pprint(grid.getStyleDirectives(GridCell('?', padding=4), 1, 1))
        {'blockBottomPadding': [('length', 4.0)],
         'blockLeftPadding': [('length', 4.0)],
         'blockRightPadding': [('length', 4.0)],
         'blockTopPadding': [('length', 4.0)]}

        >>> pprint(grid.getStyleDirectives(GridCell('?', padding=(2, 4)), 1, 1))
        {'blockBottomPadding': [('length', 4)],
         'blockLeftPadding': [('length', 2)],
         'blockRightPadding': [('length', 2)],
         'blockTopPadding': [('length', 4)]}

        >>> pprint(grid.getStyleDirectives(GridCell('?', padding=(1,2,3,4)), 1, 1))
        {'blockBottomPadding': [('length', 4)],
         'blockLeftPadding': [('length', 1)],
         'blockRightPadding': [('length', 3)],
         'blockTopPadding': [('length', 2)]}

        >>> pprint(grid.getStyleDirectives(GridCell('?', span_rows=4), 2, 5))
        {'blockSpan': [('start', '2 5'), ('stop', '2 8')]}

        >>> pprint(grid.getStyleDirectives(GridCell('?', span_cols=4), 2, 5))
        {'blockSpan': [('start', '2 5'), ('stop', '5 5')]}

        >>> pprint(grid.getStyleDirectives(GridCell('?', span_cols=3, span_rows=2), 2, 5))
        {'blockSpan': [('start', '2 5'), ('stop', '4 6')]}

    """


def test_suite():
    optionflags = (doctest.ELLIPSIS | doctest.REPORT_NDIFF |
                   doctest.NORMALIZE_WHITESPACE)
    suite = unittest.TestSuite()
    suite.addTest(doctest.DocTestSuite(optionflags=optionflags))
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2008 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
School year implementation
"""
from zope.interface import Interface
from zope.schema import Date, TextLine
from zope.location.interfaces import ILocation
from zope.container.interfaces import IWriteContainer
from zope.container.interfaces import IReadContainer
from zope.container.interfaces import IContainer
from zope.container import constraints

from schooltool.term.interfaces import ITermContainer
from schooltool.common import SchoolToolMessage as _


class IWriteSchoolYear(IWriteContainer):
    """An interface for the write aspects of a container."""


class IReadSchoolYear(IReadContainer, ILocation):

    title = TextLine(
        title=_("Title"))

    first = Date(
        title=_(u"The first day of the period of time covered."))

    last = Date(
        title=_(u"The last day of the period covered."))


class ISchoolYear(IWriteSchoolYear, IReadSchoolYear, ITermContainer):
    """School year"""


class ISchoolYearContainer(IContainer, ILocation):
    """Container for school years"""
    constraints.contains(ISchoolYear)

    def validateForOverlap(schoolyear):
        """Validate school year for overlap with other school years."""

    def getActiveSchoolYear():
        """Return the active schoolyear."""


class ISubscriber(Interface):
    """An event handler implements this"""

    def __call__():
        """Perform the action."""


class TermOverlapError(Exception):

    def __init__(self, term, overlapping_terms):
        self.term = term
        self.overlapping_terms = overlapping_terms

    def __repr__(self):
        return "Date range you have selected overlaps with term(s) (%s)" % (
            ", ".join(sorted(term.title
                             for term in self.overlapping_terms)))

    __str__ = __repr__


class TermOverflowError(Exception):

    def __init__(self, schoolyear, overflowing_terms):
        self.schoolyear = schoolyear
        self.overflowing_terms = overflowing_terms

    def __repr__(self):
        return "Date range you are trying to set is too small to contain following term(s) (%s)" % (
            ", ".join(sorted(term.title
                             for term in self.overflowing_terms)))

    __str__ = __repr__


class SchoolYearOverlapError(Exception):

    def __init__(self, schoolyear, overlapping_schoolyears):
        self.schoolyear = schoolyear
        self.overlapping_schoolyears = overlapping_schoolyears

    def __repr__(self):
        if self.schoolyear is not None:
            title = "SchoolYear '%s'" % self.schoolyear.title
        else:
            title = "DateRange"

        return  "%s overlaps with SchoolYear(s) (%s)" % (
                title,
                ", ".join(sorted(schoolyear.title
                                 for schoolyear in self.overlapping_schoolyears)))

    __str__ = __repr__



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Simple calendar events and calendars.
"""

import datetime
import random
import itertools
import email.Utils

from pytz import utc
from zope.interface import implements

from schooltool.calendar.interfaces import ICalendar, ICalendarEvent
from schooltool.calendar.mixins import CalendarEventMixin, CalendarMixin


class SimpleCalendarEvent(CalendarEventMixin):
    """A simple implementation of ICalendarEvent.

        >>> from datetime import datetime, timedelta
        >>> from zope.interface.verify import verifyObject
        >>> e = SimpleCalendarEvent(datetime(2004, 12, 15, 18, 57),
        ...                         timedelta(minutes=15),
        ...                         'Work on schooltool.calendar.simple')
        >>> verifyObject(ICalendarEvent, e)
        True

    If you do not specify a unique ID, a random one is generated

        >>> e.unique_id is not None
        True

    Additional iCal information is also supported

        >>> e2 = SimpleCalendarEvent(datetime(2004, 12, 15, 18, 57),
        ...                         timedelta(minutes=15),
        ...                         'Work on schooltool.calendar.simple',
        ...                         description="Python for fun and profit",
        ...                         location="Mt. Vernon Stable",
        ...                         recurrence='FakeRecurrance')
        >>> e2.description
        'Python for fun and profit'
        >>> e2.location
        'Mt. Vernon Stable'
        >>> e2.recurrence
        'FakeRecurrance'

    We're going to store all datetime objects in UTC

        >>> e2.dtstart.tzname()
        'UTC'

    """

    implements(ICalendarEvent)

    # Events in older versions had no allday events so this attribute
    # must be initialized in here
    allday = False

    def __init__(self, dtstart, duration, title, description=None,
                 location=None, unique_id=None, recurrence=None, allday=False):
        assert title is not None, 'title is required'

        if dtstart.tzname() not in (None, 'UTC'):
            raise ValueError, 'Can not store non UTC time info'
        self.dtstart = dtstart.replace(tzinfo=utc)

        self.duration = duration
        self.title = title
        self.description = description
        self.location = location
        self.recurrence = recurrence
        self.allday = allday
        self.unique_id = unique_id
        if not self.unique_id:
            self.unique_id = new_unique_id()


class ImmutableCalendar(CalendarMixin):
    """A simple read-only calendar.

        >>> from datetime import datetime, timedelta
        >>> from zope.interface.verify import verifyObject
        >>> e = SimpleCalendarEvent(datetime(2004, 12, 15, 18, 57),
        ...                         timedelta(minutes=15),
        ...                         'Work on schooltool.calendar.simple')
        >>> calendar = ImmutableCalendar([e])
        >>> verifyObject(ICalendar, calendar)
        True

        >>> [e.title for e in calendar]
        ['Work on schooltool.calendar.simple']

        >>> len(calendar)
        1

    """

    implements(ICalendar)

    def __init__(self, events=()):
        self._events = tuple(events)

    def __iter__(self):
        return iter(self._events)

    def __len__(self):
        return len(self._events)


def combine_calendars(*calendars):
    r"""Combine events from several calendars into one read-only calendar.

    Suppose you have several calendars with events

        >>> from datetime import datetime, timedelta
        >>> from schooltool.calendar.simple import SimpleCalendarEvent
        >>> e1 = SimpleCalendarEvent(datetime(2004, 12, 15, 18, 57),
        ...                          timedelta(minutes=15),
        ...                          'Work on schooltool.calendar.simple')
        >>> calendar1 = ImmutableCalendar([e1])
        >>> e2 = SimpleCalendarEvent(datetime(2005, 2, 2, 20, 28),
        ...                          timedelta(minutes=10),
        ...                          'Write a test for combine_calendars')
        >>> calendar2 = ImmutableCalendar([e2])

    You can combine them

        >>> calendar = combine_calendars(calendar1, calendar2)
        >>> titles = [e.title for e in calendar]
        >>> titles.sort()
        >>> print '\n'.join(titles)
        Work on schooltool.calendar.simple
        Write a test for combine_calendars

    This calendar is read-only

        >>> from schooltool.calendar.interfaces import IEditCalendar
        >>> IEditCalendar.providedBy(calendar)
        False

        >>> from schooltool.calendar.interfaces import ICalendar
        >>> ICalendar.providedBy(calendar)
        True

    """
    return ImmutableCalendar(itertools.chain(*calendars))


def new_unique_id():
    """Generate a new unique ID for a calendar event.

    UID is randomly generated and follows RFC 822 addr-spec:

        >>> uid = new_unique_id()
        >>> '@' in uid
        True

    Note that it does not have the angle brackets

        >>> '<' not in uid
        True
        >>> '>' not in uid
        True

    """
    more_uniqueness = '%d.%d' % (datetime.datetime.utcnow().microsecond,
                                 random.randrange(10 ** 6, 10 ** 7))
    # generate an rfc-822 style id and strip angle brackets
    unique_id = email.Utils.make_msgid(more_uniqueness)[1:-1]
    return unique_id




#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2008 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Tests for School year views
"""
import unittest
import doctest
from datetime import date

from zope.interface import directlyProvides
from zope.publisher.browser import TestRequest

from schooltool.term.term import Term
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.skin.skin import ISchoolToolSkin
from schooltool.schoolyear.schoolyear import SchoolYear
from schooltool.schoolyear.browser.schoolyear import SchoolYearContainerAbsoluteURLAdapter
from schooltool.schoolyear.browser.schoolyear import SchoolYearEditView
from schooltool.schoolyear.browser.schoolyear import SchoolYearAddView
from schooltool.schoolyear.interfaces import ISchoolYearContainer
from schooltool.schoolyear.testing import setUp
from schooltool.schoolyear.testing import tearDown
from schooltool.schoolyear.testing import provideStubUtility
from schooltool.schoolyear.testing import provideStubAdapter

from schooltool.schoolyear.ftesting import schoolyear_functional_layer


def doctest_SchoolYearContainerAbsoluteURLAdapter():
    """Test for absolute url adapter for school year container.

    School year container url is fixed and does not depend on the
    __name__ of the container:

        >>> syc = ISchoolYearContainer(ISchoolToolApplication(None))
        >>> str(SchoolYearContainerAbsoluteURLAdapter(syc, TestRequest()))
        'http://127.0.0.1/schoolyears'

    """


def doctest_SchoolYearAddView():
    """Tests for school year add view.

    Let's get the school year container and a request:

        >>> app = ISchoolToolApplication(None)
        >>> syc = ISchoolYearContainer(app)
        >>> request = TestRequest()
        >>> directlyProvides(request, [ISchoolToolSkin])

    Fill in the form

        >>> request.form = {'form.buttons.add'  : u'Add',
        ...                 'form.widgets.title': u'2005-2006',
        ...                 'form.widgets.first': u'2005-09-01',
        ...                 'form.widgets.last' : u'2006-07-15'}

    And add the school year

        >>> view = SchoolYearAddView(syc, request)
        >>> view.update()

        >>> view.widgets.errors
        ()

        >>> sy = syc['2005-2006']

        >>> sy.title, sy.first, sy.last
        (u'2005-2006', datetime.date(2005, 9, 1), datetime.date(2006, 7, 15))

    If we fill in the form, but click cancel - we get redirected back
    to the school year container view:

        >>> request.form = {'form.buttons.cancel'  : u'Cancel',
        ...                 'form.widgets.title': u'',
        ...                 'form.widgets.first': u'2005-09-01',
        ...                 'form.widgets.last' : u'2006-07-15'}
        >>> view = SchoolYearAddView(syc, request)
        >>> view.update()

        >>> view.request.response.getStatus()
        302
        >>> view.request.response.getHeader('Location')
        'http://127.0.0.1/schoolyears'

    """


def doctest_SchoolYearAddView_overlap():
    """Test for school year add view overlapping school year handling.

    We get the school year container

        >>> app = ISchoolToolApplication(None)
        >>> syc = ISchoolYearContainer(app)

    Add one school year to it

        >>> syc['2005-2006'] = SchoolYear("2005-2006",
        ...                               date(2005, 9, 1),
        ...                               date(2006, 7, 16))

    Fill in the form by providing a date range that overlaps the
    existing school year:

        >>> request = TestRequest()
        >>> directlyProvides(request, [ISchoolToolSkin])
        >>> request.form = {'form.buttons.add'  : u'Add',
        ...                 'form.widgets.title': u'2006-2007',
        ...                 'form.widgets.first': u'2006-07-15',
        ...                 'form.widgets.last' : u'2007-07-15'}

    And submit it

        >>> view = SchoolYearAddView(syc, request)
        >>> view.update()

    An error message should be displayed:

        >>> for error in view.widgets.errors: print error.render()
        <div class="error">
        Date range you have selected overlaps with the following school years:
        <div> <a href="http://127.0.0.1/schoolyears/2005-2006">2005-2006</a> (2005-09-01 &mdash; 2006-07-16) </div>
        </div>

    There still should be only one school year in the container:

        >>> list(syc.keys())
        [u'2005-2006']

    """


def doctest_SchoolYearEditView():
    """Tests for school year edit view.

    Let's get the school year container:

        >>> app = ISchoolToolApplication(None)
        >>> syc = ISchoolYearContainer(app)

    Add a school year:

        >>> sy = syc['2005-2006'] = SchoolYear("2005-2006",
        ...                               date(2005, 9, 1),
        ...                               date(2006, 7, 16))

    Fill in the form

        >>> request = TestRequest()
        >>> directlyProvides(request, [ISchoolToolSkin])
        >>> request.form = {'form.buttons.apply'  : u'Apply',
        ...                 'form.widgets.title': u'2006-2007',
        ...                 'form.widgets.first': u'2006-07-15',
        ...                 'form.widgets.last' : u'2007-07-15'}

    And submit it

        >>> view = SchoolYearEditView(sy, request)
        >>> view.update()

        >>> view.widgets.errors
        ()

    The school year got updated successfully:

        >>> sy.title, sy.first, sy.last
        (u'2006-2007', datetime.date(2006, 7, 15), datetime.date(2007, 7, 15))

        >>> view.status
        u'Data successfully updated.'

    We could have just canceled though

        >>> request.form = {'form.buttons.cancel': u'Cancel',
        ...                 'form.widgets.title' : u'',
        ...                 'form.widgets.first' : u'2005-09-01',
        ...                 'form.widgets.last'  : u'2006-07-15'}

    So we'd get redirected back to the container view:

        >>> view = SchoolYearEditView(sy, request)
        >>> view.update()

        >>> view.request.response.getStatus()
        302
        >>> view.request.response.getHeader('Location')
        'http://127.0.0.1/schoolyears/2005-2006'

    """


def doctest_SchoolYearEditView_shrink():
    """Test for school year editing view terms not fitting into school year.

    We get the school year container:

        >>> app = ISchoolToolApplication(None)
        >>> syc = ISchoolYearContainer(app)

    Add a school year

        >>> sy = syc['2005-2006'] = SchoolYear("2005-2006",
        ...                               date(2005, 9, 1),
        ...                               date(2006, 7, 16))

    Two terms

        >>> sy['fall'] = Term("Fall", date(2005, 9, 1), date(2006, 1, 13))
        >>> sy['spring'] = Term("Spring", date(2006, 1, 14), date(2006, 7, 16))

    Fill in the form setting new date range so one of the terms would
    not fit into the school year anymore:

        >>> request = TestRequest()
        >>> directlyProvides(request, [ISchoolToolSkin])
        >>> request.form = {'form.buttons.apply': u'Apply',
        ...                 'form.widgets.title': u'2006-2007',
        ...                 'form.widgets.first': u'2005-09-01',
        ...                 'form.widgets.last' : u'2006-07-15'}

    Then we submit the form

        >>> view = SchoolYearEditView(sy, request)
        >>> view.update()

    And get the error message displayed

        >>> for error in view.widgets.errors: print error.render()
        <div class="error">
        Date range you have selected is too small to contain the following term(s):
        <div> <a href="http://127.0.0.1/schoolyears/2005-2006/spring">Spring</a> (2006-01-14 &mdash; 2006-07-16) </div>
        </div>

    The school year should not get modified

        >>> sy.title, sy.first, sy.last
        ('2005-2006', datetime.date(2005, 9, 1), datetime.date(2006, 7, 16))

    """


def doctest_SchoolYearEditView_overlap():
    """Test for school year editing view trying to make one school year overlap another.

    We get the school year container:

        >>> app = ISchoolToolApplication(None)
        >>> syc = ISchoolYearContainer(app)

    Add a a couple of school years

        >>> syc['2005-2006'] = SchoolYear("2005-2006",
        ...                               date(2005, 9, 1),
        ...                               date(2006, 7, 16))
        >>> sy = syc['2006-2007'] = SchoolYear("2006-2007",
        ...                               date(2006, 9, 1),
        ...                               date(2007, 7, 16))

    Fill in the form so that one of the school years would overlap the
    other one

        >>> request = TestRequest()
        >>> directlyProvides(request, [ISchoolToolSkin])
        >>> request.form = {'form.buttons.apply': u'Apply',
        ...                 'form.widgets.title': u'2006-2007',
        ...                 'form.widgets.first': u'2006-07-15',
        ...                 'form.widgets.last' : u'2007-07-15'}

    And submit it

        >>> view = SchoolYearEditView(sy, request)
        >>> view.update()

    An error message will get displayed

        >>> for error in view.widgets.errors: print error.render()
        <div class="error">
        Date range you have selected overlaps with the following school years:
        <div> <a href="http://127.0.0.1/schoolyears/2005-2006">2005-2006</a> (2005-09-01 &mdash; 2006-07-16) </div>
        </div>

    And the school year will not get modified

        >>> sy.title, sy.first, sy.last
        ('2006-2007', datetime.date(2006, 9, 1), datetime.date(2007, 7, 16))

    """


def test_suite():
    optionflags = doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS
    suite = doctest.DocTestSuite(optionflags=optionflags,
                                 extraglobs={'provideAdapter': provideStubAdapter,
                                             'provideUtility': provideStubUtility},
                                 setUp=setUp, tearDown=tearDown)
    suite.layer = schoolyear_functional_layer
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2006, 2007 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Tests for SchoolTool utilities
"""
import unittest
import doctest

from zope.app.testing import setup
from zope.component.hooks import getSite


def doctest_utilityRegistration():
    """
    We set up the IntIds utility::

      >>> from schooltool.utility import UtilitySetUp
      >>> from zope.intid import IntIds
      >>> from zope.intid.interfaces import IIntIds
      >>> setup = UtilitySetUp(IntIds, IIntIds)

      >>> from schooltool.app.interfaces import CatalogSetUpEvent
      >>> site = getSite()
      >>> event = CatalogSetUpEvent(site)

    We'll pass the event to the setup::

      >>> setup(event)

    We now expect the utility to be available::

      >>> from zope.component import getUtility
      >>> util = getUtility(IIntIds)
      >>> IIntIds.providedBy(util)
      True

    """


def doctest_multipleUtilityRegistration():
    """
    A more complex scenario where multiple dependent utilities are
    registered, the intids utility and the catalog. The catalog also
    contains indexes, so need special setup code to initialize the indexes.

    Let's import the information needed to set up the IntIds utility first::

      >>> from zope.intid import IntIds
      >>> from zope.intid.interfaces import IIntIds

    And now for the catalog::

      >>> from zope.catalog.catalog import Catalog
      >>> from zope.catalog.interfaces import ICatalog
      >>> from zope.catalog.field import FieldIndex

    We'll define the catalog setup function that creates the index::

      >>> from zope.interface import Interface
      >>> def catalogSetUp(catalog):
      ...     catalog['myindex'] = FieldIndex(Interface, 'foo')

    Now let's set these up as utilities::

      >>> from schooltool.utility import UtilitySpecification
      >>> from schooltool.utility import MultiUtilitySetUp
      >>> setup = MultiUtilitySetUp(
      ...    UtilitySpecification(IntIds, IIntIds),
      ...    UtilitySpecification(Catalog, ICatalog, setUp=catalogSetUp))

      >>> from schooltool.app.interfaces import CatalogSetUpEvent
      >>> site = getSite()
      >>> event = CatalogSetUpEvent(site)
      >>> setup(event)

    We'll check whether the utilities are available::

      >>> from zope.component import getUtility
      >>> intids = getUtility(IIntIds)
      >>> IIntIds.providedBy(intids)
      True
      >>> catalog = getUtility(ICatalog)
      >>> ICatalog.providedBy(catalog)
      True

    """


def doctest_utilityOverride():
    """
    Let's create a utility::

      >>> from persistent import Persistent
      >>> from zope.container.contained import Contained
      >>> from zope.interface import Interface, implements
      >>> class IMyUtility(Interface):
      ...     pass
      >>> class MyUtility1(Persistent, Contained):
      ...     implements(IMyUtility)

    Let's register it::

      >>> from schooltool.app.interfaces import CatalogSetUpEvent
      >>> from schooltool.utility import UtilitySetUp
      >>> setup = UtilitySetUp(MyUtility1, IMyUtility)
      >>> from zope.component.hooks import getSite
      >>> site = getSite()
      >>> event = CatalogSetUpEvent(site)
      >>> setup(event)

    And let's register another one, overriding the first::

      >>> class MyUtility2(Persistent, Contained):
      ...   implements(IMyUtility)
      >>> setup2 = UtilitySetUp(MyUtility2, IMyUtility, override=True)
      >>> setup2(event)

    Now we expect there to be only MyUtility2::

      >>> from zope.component import getUtility
      >>> util = getUtility(IMyUtility)
      >>> isinstance(util, MyUtility2)
      True

    """


def sitePlacefulSetUp(*args):
    setup.placefulSetUp(site=True)

def sitePlacefulTearDown(*args):
    setup.placefulTearDown()

def test_suite():
    return unittest.TestSuite([
        doctest.DocTestSuite(setUp=sitePlacefulSetUp,
                             tearDown=sitePlacefulTearDown)])



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool flourish forms.
"""
from zope.component import adapts
from zope.interface import Interface
from zope.schema.interfaces import IBool
from zope.schema import ValidationError
import z3c.form.form
from z3c.form.error import ErrorViewSnippet
from z3c.form.term import BoolTerms
from z3c.form.interfaces import IRadioWidget

from schooltool.skin.flourish import page, content, viewlet, tal
from schooltool.common.inlinept import InheritTemplate
from schooltool.skin.flourish.interfaces import IFlourishLayer

from schooltool.common import SchoolToolMessage as _


class FlourishRadioBoolTerms(BoolTerms):

    adapts(Interface,
           IFlourishLayer,
           Interface,
           IBool,
           IRadioWidget)

    trueLabel = _('Yes')
    falseLabel = _('No')


class Form(z3c.form.form.Form, page.Page):
    __call__ = page.Page.__call__
    formErrorsMessage = _('Please correct the marked fields below.')
    template = InheritTemplate(page.Page.template)

    def update(self):
        super(Form, self).update()

    def updateActions(self):
        super(Form, self).updateActions()


class DisplayForm(z3c.form.form.DisplayForm, page.Page):
    template = InheritTemplate(page.Page.template)
    __call__ = page.Page.__call__

    def update(self):
        super(DisplayForm, self).update()


class AddForm(Form, z3c.form.form.AddForm):

    def update(self):
        super(AddForm, self).update()
        if self._finishedAdd:
            self.request.response.redirect(self.nextURL())
            return ""

    @z3c.form.button.buttonAndHandler(_('Add'), name='add')
    def handleAdd(self, action):
        return z3c.form.form.AddForm.handleAdd.func(self, action)

    @z3c.form.button.buttonAndHandler(_('Cancel'), name='cancel')
    def handleCancel(self, action):
        self.request.response.redirect(self.nextURL())


class Dialog(page.Page):

    # Set this to False if you don't want the browser to reload the whole
    # page, but display the redirected result in the dialog instead.
    reload_parent = True

    # These will be passed to the dialog handler.
    ajax_settings = None

    def __init__(self, *args, **kw):
        super(Dialog, self).__init__(*args, **kw)
        self.ajax_settings = {}

    def initDialog(self):
        self.ajax_settings['dialog'] = {
            'autoOpen': True,
            'modal': True,
            'resizable': False,
            'draggable': False,
            'position': ['center','middle'],
            'width': '560',
            }

    def updateDialog(self):
        pass

    def update(self):
        self.initDialog()
        super(Dialog, self).update()
        self.updateDialog()

    def __call__(self, *args, **kw):
        result = super(Dialog, self).__call__(*args, **kw)

        if not self.ajax_settings.get('html'):
            self.ajax_settings['html'] = result

        response = self.request.response
        if (self.reload_parent and
            not self.ajax_settings.get('redirect') and
            response.getStatus() in [300, 301, 302, 303, 304, 305, 307]):
            self.ajax_settings['redirect'] = response.getHeader('Location')
            self.ajax_settings['dialog'] = 'close'

        response.reset()
        response.setHeader('Content-Type', 'application/json')
        encoder = tal.JSONEncoder()

        json = encoder.encode(self.ajax_settings)
        return json


class DialogForm(Dialog, Form):
    dialog_close_actions = ()
    dialog_submit_actions = ()

    def updateActions(self):
        super(DialogForm, self).updateActions()
        for name in self.dialog_submit_actions:
            action = self.actions[name]
            action.onclick = u'return ST.dialogs.submit(this, this);'
        for name in self.dialog_close_actions:
            action = self.actions[name]
            action.onclick = u'return ST.dialogs.close(this);'


class DialogAddForm(DialogForm, AddForm):
    pass


class FlourishErrorViewSnippet(ErrorViewSnippet):

    adapts(ValidationError, IFlourishLayer, None, None, None, None)

    def update(self):
        super(FlourishErrorViewSnippet, self).update()
        self.widget.addClass('error')


class FormContent(content.ContentProvider, z3c.form.form.Form):

    def update(self):
        z3c.form.form.Form.update(self)
        content.ContentProvider.update(self)

    def render(self, *args, **kw):
        return z3c.form.form.Form.render(self)


class FormViewlet(viewlet.Viewlet, z3c.form.form.Form):

    def update(self):
        z3c.form.form.Form.update(self)
        viewlet.Viewlet.update(self)

    def render(self, *args, **kw):
        return z3c.form.form.Form.render(self)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2007 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Basic person sample data generation
"""
import datetime
import os
from pytz import utc

from zope.interface import implements
from zope.security.proxy import removeSecurityProxy

from schooltool.basicperson.person import BasicPerson
from schooltool.sampledata import PortableRandom
from schooltool.sampledata.interfaces import ISampleDataPlugin
from schooltool.sampledata.name import NameGenerator
from schooltool.group.interfaces import IGroupContainer
from schooltool.term.interfaces import ITermContainer
from schooltool.app.interfaces import ISchoolToolCalendar
from schooltool.app.cal import CalendarEvent
from schooltool.common import DateRange


class ChoiceGenerator(object):
    def __init__(self, seed, choices):
        self.random = PortableRandom(seed)
        self.choices = choices

    def generate(self):
        return self.random.choice(self.choices)


class SampleStudents(object):

    implements(ISampleDataPlugin)

    name = 'students'
    dependencies = ('terms', )

    # Number of persons to generate
    power = 1000

    def personFactory(self, namegen, prefixgen, gendergen, count):
        first_name, last_name, full_name = namegen.generate()
        person_id = 'student%03d' % count
        person = BasicPerson(person_id, first_name, last_name)
        person.setPassword(person_id)
        person.gender = gendergen.generate()
        return person

    def generate(self, app, seed=None):
        namegen = NameGenerator(str(seed) + self.name)
        prefixgen = ChoiceGenerator(str(seed), ['Mr', 'Mrs', 'Miss', ''])
        gendergen = ChoiceGenerator(str(seed), ['male', 'female'])

        students = IGroupContainer(app)['students']
        for count in range(self.power):
            person = self.personFactory(namegen, prefixgen, gendergen, count)
            # Without removeSecurityProxy we can't add members a
            # group.
            removeSecurityProxy(students.members).add(person)
            app['persons'][person.__name__] = person


class SampleTeachers(object):
    implements(ISampleDataPlugin)

    name = 'teachers'
    dependencies = ('terms', )

    # Number of teachers to generate
    power = 48

    def personFactory(self, namegen, count):
        first_name, last_name, full_name = namegen.generate()
        person_id = 'teacher%03d' % count
        person = BasicPerson(person_id, first_name, last_name)
        person.setPassword(person_id)
        return person

    def generate(self, app, seed=None):
        namegen = NameGenerator(str(seed) + self.name)
        teachers = IGroupContainer(app)['teachers']
        for count in range(self.power):
            person = self.personFactory(namegen, count)
            # Without removeSecurityProxy we can't add members a
            # group.
            removeSecurityProxy(teachers.members).add(person)
            app['persons'][person.__name__] = person


class SamplePersonalEvents(object):
    """Sample data plugin class that generates personal random events."""

    implements(ISampleDataPlugin)

    name = 'personal_events'
    dependencies = ('students', 'teachers', 'terms')

    probability = 2     # probability of having an event on any day

    def _readLines(self, filename):
        """Read in lines from file

        Filename is relative to the module.
        Returned lines are stripped.
        """
        fullpath = os.path.join(os.path.dirname(__file__), filename)
        lines = file(fullpath).readlines()
        return [line.strip() for line in lines]

    def generate(self, app, seed=None):
        self.random = PortableRandom(seed)
        event_titles = self._readLines('event_titles.txt')
        person_ids = [person for person in app['persons'].keys()
                      if person.startswith('student') or
                         person.startswith('teacher')]
        dates = []
        for term in ITermContainer(app).values():
            dates.append(term.first)
            dates.append(term.last)
        first = min(dates)
        last = max(dates)
        days = DateRange(first, last)
        for person_id in person_ids:
            person = app['persons'][person_id]
            calendar = ISchoolToolCalendar(person)
            for day in days:
                if self.random.randrange(0, 100) < self.probability:
                    event_title = self.random.choice(event_titles)
                    time_hour = self.random.randint(6, 23)
                    time_min = self.random.choice((0, 30))
                    event_time = datetime.datetime(day.year,
                                                   day.month,
                                                   day.day,
                                                   time_hour,
                                                   time_min,
                                                   tzinfo=utc)
                    event_duration = datetime.timedelta(
                                       minutes=self.random.randint(1, 12)*30)
                    event = CalendarEvent(event_time,
                                          event_duration,
                                          event_title)
                    calendar.addEvent(event)




#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Selenium testing.
"""
from __future__ import absolute_import
import asyncore
import cgi
import codecs
import doctest
import linecache
import os
import re
import string
import sys
import threading
import time
import urllib
import unittest
import xlrd
from StringIO import StringIO
from UserDict import DictMixin

import lxml.html
import lxml.doctestcompare
import lxml.etree

from zope.app.wsgi import WSGIPublisherApplication
from zope.server.taskthreads import ThreadedTaskDispatcher
from zope.server.http.commonaccesslogger import CommonAccessLogger
from zope.server.http.wsgihttpserver import WSGIHTTPServer


import schooltool.testing.registry
from schooltool.testing import mock
from schooltool.testing.functional import ZCMLLayer
from schooltool.testing.functional import collect_txt_ftests


BLACK_MAGIC = True # enable morally ambiguous monkeypatching

_browser_factory = None
selenium_enabled = False

IMPLICIT_WAIT = 5 # in seconds

try:
    if _browser_factory is None:
        import schooltool.devtools.selenium_recipe
        _browser_factory = schooltool.devtools.selenium_recipe.spawn_browser
        selenium_enabled = (
            schooltool.devtools.selenium_recipe.default_factory is not None)
        IMPLICIT_WAIT = schooltool.devtools.selenium_recipe.default_browser_config.implicit_wait
except ImportError:
    pass


try:
    from selenium.webdriver.remote.webelement import WebElement \
            as SimpleWebElement

    if _browser_factory is None:
        import selenium.webdriver.firefox.webdriver
        def _browser_factory(factory_name='firefox', config=None):
            assert factory_name == 'firefox'
            return selenium.webdriver.firefox.webdriver.WebDriver()
        selenium_enabled = True

    from selenium.common.exceptions import NoSuchElementException
    from selenium.common.exceptions import TimeoutException
except ImportError:
    SimpleWebElement = object
    pass


_Browser_UI_ext = None
_Element_UI_ext = None


class ExtensionGroup(object):
    def __init__(self):
        self._handlers = {}

    def __repr__(self):
        return '<ExtensionGroup (%d ext)>: %s' % (
            len(self._handlers), list(self._handlers))

    def __getitem__(self, name):
        if name not in self:
            raise KeyError(name)
        return self._handlers[name]

    def __iter__(self):
        return iter(self._handlers)

    def __setitem__(self, name, handler):
        if name in self._handlers:
            raise KeyError(name)
        self._handlers[name] = handler

    def __delitem__(self, name, handler):
        del self._handlers[name]

    def __contains__(self, name):
        return name in self._handlers


class BoundExtension(object):

    _extension = None
    _name = None
    _target = None

    def __init__(self, extension, name, target):
        self._extension = extension
        self._name = name
        self._target = target
        if isinstance(extension, ExtensionGroup):
            for ext_n in extension:
                # Lazy man's dynamic attr marker
                setattr(self, ext_n, None)

    def __getattribute__(self, name):
        extension = object.__getattribute__(self, '_extension')
        if (name in BoundExtension.__dict__ or
            name not in extension):
            return object.__getattribute__(self, name)
        return BoundExtension(
                extension[name],
                '%s.%s' % (object.__getattribute__(self, '_name'), name),
                object.__getattribute__(self, '_target'))

    def __call__(self, *args, **kw):
        return self._extension(self._target, *args, **kw)

    def __repr__(self):
        return '<BoundExtension %s (%r)>' % (self._name, self._target)


def registerExtension(target, name, handler):
    """Register UI extension for web elements."""
    path = filter(None, [s.strip() for s in name.split('.')])
    if not path:
        raise ValueError(name)
    while len(path) > 1:
        part = path.pop(0)
        if part not in target:
            target[part] = ExtensionGroup()
        target = target[part]
    target[path[0]] = handler


def registerBrowserUI(name, handler):
    registerExtension(_Browser_UI_ext, name, handler)


def registerElementUI(name, handler):
    registerExtension(_Element_UI_ext, name, handler)


def extractJSONWireParams(cmd):
    t = string.Template(cmd)
    return [named
            for (excaped, named, braced, invalid)
            in t.pattern.findall(cmd)]


def extractJSONWireCommands(remote_connection):
    # XXX: hacky.
    for cmd, info in sorted(remote_connection._commands.items()):
        method, path = info
        params = extractJSONWireParams(path)
        yield cmd, params


class CommandExecutor(object):
    driver = None
    _commands = None

    def __init__(self, driver, browser=None):
        self.driver = driver
        self.browser = browser
        self._commands = {}

    def extractDriverCommands(self):
        """Extract commands / params from driver's command executor."""
        signatures = extractJSONWireCommands(self.driver.command_executor)
        for name, params in signatures:
            self.addDriverCommand(name, params)

    def addDriverCommand(self, name, params):
        """Add a method to this instance that executes corresponding
        command in the web driver."""
        def command(**keywords):
            return self.driver.execute(name, keywords)
        command.__doc__ = \
            '''%s(%s, ...)''' % (
            name, ', '.join(['%s=?' % p for p in params]))
        self._commands[name] = command

    def __getattr__(self, name):
        if name in self._commands:
            return self._commands[name]
        return object.__getattr__(self, name)

    def printDriverCommands(self):
        """Print configured commands."""
        for name in sorted(self._commands):
            print self._commands[name].__doc__


class SortedDict(DictMixin):

    def __init__(self, dict):
        self.dict = dict

    def __getitem__(self, k):
        return self.dict.__getitem__(k)

    def __setitem__(self, k, v):
        return self.dict.__setitem__(k)

    def __delitem__(self, k):
        return self.dict.__delitem__(k)

    def keys(self):
        return sorted(self.dict)


class DocTestHtmlElement(object):

    @property
    def attrib(self):
        return SortedDict(super(DocTestHtmlElement, self).attrib)

    def keys(self):
        u"""keys(self)

        Gets a list of attribute names.  The names are returned in an
        arbitrary order (just like for an ordinary Python dictionary).
        """
        return sorted(super(DocTestHtmlElement, self).keys())

    def values(self):
        u"""values(self)

        Gets element attribute values as a sequence of strings.  The
        attributes are returned in an arbitrary order.
        """
        return sorted(super(DocTestHtmlElement, self).values())

    def items(self):
        u"""items(self)
        Gets element attributes, as a sequence. The attributes are returned in
        an arbitrary order.
        """
        return sorted(super(DocTestHtmlElement, self).items())


class HtmlElement(DocTestHtmlElement, lxml.html.HtmlElement):
    pass


class DocTestHtmlElementClassLookup(lxml.html.HtmlElementClassLookup):

    def lookup(self, node_type, document, namespace, name):
        if node_type == 'element':
            return self._element_classes.get(name.lower(), HtmlElement)
        return lxml.html.HtmlElementClassLookup.lookup(
            self, node_type, document, namespace, name)


class DocTestHTMLParser(lxml.etree.HTMLParser):
    def __init__(self, **kwargs):
        super(DocTestHTMLParser, self).__init__(**kwargs)
        self.set_element_class_lookup(
            DocTestHtmlElementClassLookup(
                mixins=[('*', DocTestHtmlElement)],
                ))


class DocTestXHTMLParser(lxml.etree.XMLParser):
    def __init__(self, **kwargs):
        super(DocTestHTMLParser, self).__init__(**kwargs)
        self.set_element_class_lookup(
            DocTestHtmlElementClassLookup(
                mixins=[('*', DocTestHtmlElement)],
                ))


shared_html_parser = DocTestHTMLParser(
    recover=True, remove_blank_text=False)

shared_xhtml_parser = DocTestHTMLParser(
    recover=True, remove_blank_text=False)


class HTMLSerializer(object):

    container_node = None

    empty_tags = (
        'param', 'img', 'area', 'br', 'basefont', 'input',
        'base', 'meta', 'link', 'col')

    def __init__(self, doc=None, parser=shared_html_parser):
        self.parser = parser
        if isinstance(doc, basestring):
            self.doc = self.parse(doc)
        elif isinstance(doc, lxml.etree.ElementBase):
            self.doc = doc
        else:
            raise NotImplemented('doc must be html string or lxml element,'
                                 ' got: %r' % doc)

    def _fix_ns(self, text):
        # Do some guesswork to remove the namespace for now.
        # XXX: regex or something would be better
        if isinstance(text, basestring):
            if text[0] == '{':
                return text.split('}')[-1]
        return text

    def parse(self, html, **kw):
        self.container_node = None
        start = html[:200].lstrip().lower()
        inject_html = bool(not '<html>' in start)
        inject_body = bool(not '<body>' in start)
        template = '%s'
        if inject_body:
            template = '<body>' + template + '</body>'
        if inject_html:
            template = '<html>' + template + '</html>'
        if template != '%s':
            html = template % html
        doc = lxml.etree.fromstring(html, parser=self.parser, **kw)

        if inject_body:
            bodies = doc.xpath('//body')
            assert len(bodies) == 1, bodies
            self.container_node = bodies[0]
        elif inject_html:
            htmls = doc.xpath('//html')
            assert len(htmls) == 1, htmls
            self.container_node = htmls[0]

    def format_doc(self, indent=0):
        stream = StringIO()
        self.print_doc(stream=stream, indent=indent)
        return stream.getvalue()

    def print_doc(self, stream=sys.stdout, indent=0):
        if self.container_node is None:
            self.print_node(self.doc, stream=stream, indent=indent)
        else:
            for node in self.container_node:
                self.print_node(node, stream=stream, indent=indent)

    def print_node(self, node, stream=sys.stdout, indent=0):
        for base, method in self.formatters:
            if isinstance(node, base):
                return method(self, node, stream=stream, indent=indent)
        raise NotImplemented(repr(node))

    def print_text(self, text, stream=sys.stdout, indent=0):
        if text is None:
            return
        text = text.strip()
        if not text:
            return
        text = (indent*' ' + '\n').join(text.splitlines())
        stream.write(indent*' '+cgi.escape(text, True)+'\n')

    def print_comment(self, node, stream=sys.stdout, indent=0):
        stream.write(indent*' ' + '<!--\n')
        self.print_text(node.text, stream=stream, indent=indent+2)
        stream.write(indent*' ' + '-->\n')

    def print_tag(self, node, stream=sys.stdout, indent=0):
        has_contents = bool(len(node) or node.text)
        # Only auto-close node if it correctly has no children or text
        if (self._fix_ns(node.tag).lower() in self.empty_tags and
            not has_contents):
            self.auto_tag(node, stream=stream, indent=indent)
        else:
            self.open_tag(node, stream=stream, indent=indent)
            self.print_text(node.text, stream=stream, indent=indent+2)
            for child in node:
                self.print_node(child, stream=stream, indent=indent+2)
            self.close_tag(node, stream=stream, indent=indent)
            self.print_text(node.tail, stream=stream, indent=indent)

    def print_attribs(self, node, stream=sys.stdout):
        process = lambda t: t and cgi.escape(unicode(t), True) or ""
        stream.write(' ' .join(['%s="%s"' % (self._fix_ns(n), process(v))
                                for n, v in node.attrib.items()]))

    def auto_tag(self, node, stream=sys.stdout, indent=0):
        stream.write(indent*' ' + '<'+self._fix_ns(node.tag).lower())
        if node.attrib:
            stream.write(' ')
            self.print_attribs(node, stream=stream)
        stream.write(' />\n')

    def open_tag(self, node, stream=sys.stdout, indent=0):
        stream.write(indent*' ' + '<'+self._fix_ns(node.tag).lower())
        if node.attrib:
            stream.write(' ')
            self.print_attribs(node, stream=stream)
        stream.write('>\n')

    def close_tag(self, node, stream=sys.stdout, indent=0):
        stream.write(indent*' ' + '</' + self._fix_ns(node.tag).lower() + '>\n')

    def skip(self, *args, **kw):
        pass

    formatters = (
            (lxml.etree.CommentBase, print_comment),
            (lxml.etree.ElementBase, print_tag),
            (lxml.etree.PIBase, skip),
            )


class PrintablesList(list):

    def __unicode__(self):
        return '\n'.join([unicode(el).rstrip() for el in self])

    __str__ = lambda self: unicode(self).encode('utf-8')


class ForwardingMethodsList(PrintablesList):

    fwd_methods = ()
    fwd_attrs = ()

    def __getattr__(self, name):
        if (name in self.fwd_methods or
            name in self.fwd_attrs):
            attrs = []
            for web_el in self:
                method = getattr(web_el, name, None)
                if method is None:
                    attrs.append(None)
                    continue
                if hasattr(method, '__call__'):
                    attrs.append(method)
                else:
                    attrs.append(method)
            if name in self.fwd_attrs:
                return PrintablesList(attrs)
            def multi_element_method(*args, **kw):
                results = []
                for attr in attrs:
                    if hasattr(method, '__call__'):
                        results.append(attr(*args, **kw))
                    else:
                        results.append(attr)
                return PrintablesList(results)
            return multi_element_method
        return PrintablesList.__getattr__(self, name)


class WebElementList(ForwardingMethodsList):

    fwd_methods = (
        'get_attribute', 'is_selected', 'is_enabled',
        'value_of_css_property',
        )

    fwd_attrs = (
        'tag_name', 'text', 'size', 'location',
        )


def sanitizeHTML(html):
    if html.strip():
        serializer = HTMLSerializer(html)
        return serializer.format_doc()
    return html


def getWebElementHTML(driver, web_elements):
    """Get HTML of WebElement or a list of WebElement."""
    if isinstance(web_elements, list):
        snippets = []
        for el in web_elements:
            snippets.append(getWebElementHTML(driver, el))
        html = u'\n'.join(snippets)
    else:
        html = driver.execute_script(
            'return document.createElement("div")'
            '.appendChild(arguments[0].cloneNode(true))'
            '.parentNode.innerHTML',
            web_elements)
    if html:
        html = sanitizeHTML(html)
    return html


class WebElement(SimpleWebElement):

    browser = None
    query = None
    query_all = None

    def __init__(self, *args):
        browser = None
        if len(args) == 1:
            web_element = args[0]
            assert isinstance(web_element, SimpleWebElement)
            parent, id_ = web_element.parent, web_element.id
        elif len(args) == 2:
            parent, id_ = args
        elif len(args) == 3:
            parent, id_, browser = args
        else:
            raise TypeError(
                "__init__() takes 1 (web_element) or 2 (parent, id) arguments"
                "(%d given)" % len(args))
        SimpleWebElement.__init__(self, parent, id_)
        self.browser = browser
        self.query = WebElementQuery(self, single=True, browser=browser)
        self.query_all = WebElementQuery(self, single=False, browser=browser)

    @property
    def expired(self):
        try:
            from selenium.common.exceptions import StaleElementReferenceException
            self.tag_name
        except StaleElementReferenceException:
            return True
        return False

    @property
    def ui(self):
        return BoundExtension(_Element_UI_ext, 'element.ui', self)

    def type(self, *value):
        return self.send_keys(*value)

    def getHTML(self):
        """Get HTML of WebElement."""
        return getWebElementHTML(self.parent, self)

    def __unicode__(self):
        return self.getHTML()

    __str__ = lambda self: unicode(self).encode('utf-8')


def proxy_find_element(name):
    def find_element(self, param):
        if self._single:
            method = getattr(self._target, "find_element_by_%s" % name)
        else:
            method = getattr(self._target, "find_elements_by_%s" % name)
        return self._wrap(method(param))
    return find_element


class WebElementQuery(object):

    _single = False
    _target = None
    _browser = None

    def __init__(self, target, single=False, browser=None):
        self._target = target
        self._single = single
        self._browser = browser

    def _wrap(self, web_element):
        if isinstance(web_element, WebElement):
            return web_element # already wrapped
        elif isinstance(web_element, SimpleWebElement):
            return WebElement(web_element.parent, web_element.id, self._browser)
        elif isinstance(web_element, dict):
            return dict([(key, self._wrap(val))
                         for key, val in web_element.items()])
        elif isinstance(web_element, list):
            return WebElementList([self._wrap(el) for el in web_element])
        else:
            return web_element

    _link_text = proxy_find_element("link_text")
    _partial_link_text = proxy_find_element("partial_link_text")

    name = proxy_find_element("name")
    id = proxy_find_element("id")
    css = proxy_find_element("css_selector")
    tag = proxy_find_element("tag_name")
    xpath = proxy_find_element("xpath")

    def link(self, text=None, url=None, partial=False):
        if not ((url is not None) ^ (text is not None)):
            raise Exception("Must specify either url or link text")
        if partial:
            if url is not None:
                return self.xpath('//a[@href~="%s"]' % url)
            return self._partial_link_text(text)
        else:
            if url is not None:
                return self.xpath('//a[@href="%s"]' % url)
            return self._link_text(text)

    def button(self, text=None, name=None):
        if not ((text is not None) ^ (name is not None)):
            raise Exception("Must specify either button text or value")
        if name is not None:
            return self.xpath(
                '//input[@type="submit" and @name="%s"]' % name)
        return self.xpath(
            '//input[@type="submit" and contains(@value, "%s")]' % text)

    @property
    def active(self):
        """Return an active element."""
        import selenium.webdriver.remote.webdriver
        target = self._target
        while not isinstance(
            target, selenium.webdriver.remote.webdriver.WebDriver):
            target = target.parent
        active = target.switch_to_active_element()
        if not self._single:
            active = [active]
        return self._wrap(active)

    def form(self, some_selectors=None):
        # XXX: return home-made wrapped form filler(s)
        raise NotImplemented()


class FileLink(object):

    path = None
    url = None
    message = None

    def __init__(self, full_path, url=None, message=None):
        self.path = os.path.normpath(full_path)
        self.url = url
        self.message = message

    def __nonzero__(self):
        return os.path.exists(self.path)

    @property
    def filename(self):
        return os.path.basename(self.path)

    def __repr__(self):
        url = self.url or self.path
        return '<%s %s%s%s>' % (
            self.__class__.__name__, self.filename,
            self.message and (' ' + repr(self.message)) or '',
            url and (' ' + url + ' ') or '')


class Download(FileLink):

    @property
    def file_format(self):
        fname = self.filename.lower()
        if fname.endswith('xls'):
            return 'xls'
        else:
            return 'txt'

    def read(self, format=None, encoding=None):
        if format is None:
            format = self.file_format
        if format == 'xls':
            return xlrd.open_workbook(self.path)
        elif format == 'txt':
            f = codecs.open(self.path, encoding=encoding)
            text = f.read()
            f.close()
            return text
        else:
            raise NotImplemented('Reading file format %r' % format)


class Screenshot(FileLink):
    pass


class Browser(object):

    WAIT_FEW_SECONDS = max(5, min(IMPLICIT_WAIT, 30))
    WAIT_FEW_MINUTES = max(120, min(IMPLICIT_WAIT*12, 600))
    WAIT_LONG = max(300, min(IMPLICIT_WAIT*80, 3600))

    driver = None
    execute = None
    query = None
    query_all = None

    # helpers
    keys = None

    def __init__(self, pool, driver):
        self.pool = pool
        self.driver = driver
        self.execute = CommandExecutor(self.driver)
        self.execute.extractDriverCommands()
        self.query = WebElementQuery(
            self.driver, single=True, browser=self)
        self.query_all = WebElementQuery(
            self.driver, single=False, browser=self)
        import selenium.webdriver.common.keys
        self.keys = selenium.webdriver.common.keys.Keys()

    def open(self, url="http://localhost/"):
        """Open an URL."""
        server_url = self.pool.localhost[:-1]
        url = url.replace('http://localhost', server_url)
        return self.driver.get(url)

    def close(self):
        """Close the browser."""
        return self.driver.quit()

    def type(self, *keys):
        el = self.driver.switch_to_active_element()
        return el.send_keys(*keys)

    def execute_script(self, script, *args, **kw):
        async = kw.get('async', False)
        # XXX: missing:
        raise NotImplemented()

    def screenshot(self, name='screenshot', ext='png', overwrite=False):
        layer = self.pool.layer

        if os.path.isabs(name):
            name = os.path.normpath(name)
            directory = os.path.dirname(name)
            name = os.path.basename(name)
        else:
            directory = getattr(layer.config, 'screenshots_dir', None)

        if directory is None:
            return None

        name = name.strip()
        if name.lower().endswith(('.'+ext).lower()):
            name = name[:-(len('.'+ext))]

        fname = name+'.'+ext

        full_name = os.path.join(directory, fname)
        n = 1
        while (os.path.exists(full_name) and not overwrite):
            n += 1
            fname = '%s-%d.%s' % (name, n, ext)
            full_name = os.path.join(directory, fname)

        if overwrite and os.path.exists(full_name):
            os.remove(full_name)

        try:
            result = self.driver.save_screenshot(full_name)
        except Exception, error:
            return Screenshot(full_name, message=error)

        if not result:
            return Screenshot(
                full_name, message="Failed to save screenshot")

        url = None
        if getattr(layer.config, 'screenshots_url', None):
            url = urllib.basejoin(layer.config.screenshots_url, fname)

        return Screenshot(full_name, url=url)

    # XXX: missing:
    #  - window handles, and more

    @property
    def url(self):
        return self.driver.current_url

    @property
    def title(self):
        return self.driver.title

    @property
    def contents(self):
        return sanitizeHTML(self.driver.page_source)

    def printHTML(self, web_element):
        """Print HTML of WebElement or a list of WebElement."""
        print self.getHTML(web_element)

    def getHTML(self, web_element):
        """Get HTML of WebElement or a list of WebElement."""
        return getWebElementHTML(self.driver, web_element)

    @property
    def ui(self):
        return BoundExtension(_Browser_UI_ext, 'browser.ui', self)

    def wait(self, checker, wait=None, no_element_result=False):
        if wait is None:
            wait = self.WAIT_FEW_SECONDS
        start = time.time()
        ts = 0.3
        while True:
            try:
                result = checker()
            except NoSuchElementException:
                result = no_element_result
            if result:
                break
            now = time.time()
            if now >= start+wait:
                raise TimeoutException()
            time.sleep(ts)
            ts = min(50, ts*1.01)

    def wait_no(self, checker, wait=None):
        return self.wait(lambda: not checker(), wait=wait, no_element_result=True)

    def wait_few_minutes(self, checker):
        return self.wait(checker, wait=self.WAIT_FEW_MINUTES)

    def wait_long(self, checker):
        return self.wait(checker, wait=self.WAIT_LONG)

    def wait_page(self, action):
        page = self.query.tag('html')
        result = action()
        self.wait(lambda: page.expired)
        return result


class Downloads(object):

    def __init__(self, layer):
        self.layer = layer
        self.reset()

    @property
    def path(self):
        if self.layer.config is None:
            return None
        return self.layer.config.downloads_dir

    def _created(self, fname):
        tstamp = os.path.getctime(os.path.join(self.path, fname))
        return tstamp

    def _size(self, fname):
        size = os.path.getsize(os.path.join(self.path, fname))
        return size

    def _wrap(self, fname):
        url = None
        if getattr(self.layer.config, 'downloads_url', None):
            url = urllib.basejoin(self.layer.config.downloads_url, fname)
        return Download(
            os.path.join(self.path, fname),
            url=url, message=None)

    def update(self):
        """Update the list of downloads."""
        if self.path is None:
            return
        for name in os.listdir(self.path):
            # unfinished download
            if name.endswith('.crdownload') or name.endswith('.part'):
                continue
            # empty file
            if not self._size(name):
                continue
            timestamp = self._created(name)
            if (name not in self.files):
                # file is new:
                self.files[name] = (self._created(name), 'N')
            elif self.files[name][0] < timestamp:
                # Float comparison ^ makes me uneasy
                # file was re-created:
                self.files[name] = (self._created(name), 'N')

    def reset(self):
        """Clear the list of downloads (does not delete actual files)"""
        self.files = {}
        if self.path is None:
            return
        assert os.path.isdir(self.path)
        for name in os.listdir(self.path):
            self.files[name] = ('I', self._created(name))

    def clear(self):
        """Clear downloads"""
        self.files = {}
        if self.path is None:
            return
        for name in os.listdir(self.path):
            os.remove(os.path.join(self.path, name))

    def get(self, name=None):
        """Return the [latest by default] download"""
        if self.path is None:
            return None
        if name is None:
            new = self.new()
            if not new:
                return None
            name = new[0].filename
        else:
            self.update()

        if (name not in self.files or
            self.files[name][1] == 'I'):
            return None
        self.files[name] = (self.files[name][0], 'O')

        return self._wrap(name)

    def read(self, name=None, format=None, encoding=None):
        """Return the contents of the [latest by default] download"""
        download = self.get(name=name)
        if download is None:
            return None
        return download.read(format=format, encoding=encoding)

    def new(self):
        """List the new, "unread" downloads."""
        if self.path is None:
            return []
        self.update()
        sorted_files = sorted(self.files.items(), key=lambda i: (i[1], i[0]))
        result = [name
                  for name, status in sorted_files
                  if status[1] == 'N']
        return [self._wrap(f) for f in sorted(result)]

    def all(self):
        """List all downloads."""
        if self.path is None:
            return []
        self.update()
        sorted_files = sorted(self.files.items(), key=lambda i: (i[1], i[0]))
        result = [name
                  for name, status in sorted_files
                  if status[0] != 'I']
        return [self._wrap(f) for f in sorted(result)]


class BrowserPool(object):

    layer = None
    browsers = None

    def __init__(self, layer):
        self.layer = layer
        self.browsers = {}

    def __getattr__(self, name):
        if (name.startswith('_') or
            name == 'trait_names'): # for IPdb.
            raise AttributeError(name)
        if name not in self.browsers:
            return self.start(name)
        return self.browsers[name]

    def start(self, name):
        if name in self.browsers:
            self.quit(name)
        driver = _browser_factory(config=self.layer.config)
        self.browsers[name] = Browser(self, driver)
        return self.browsers[name]

    def quit(self, name):
        self.browsers[name].close()
        del self.browsers[name]

    def reset(self):
        for name in list(self.browsers):
            self.quit(name)

    @property
    def localhost(self):
        assert self.layer.serving
        return 'http://%s:%s/' % self.layer.server.socket.getsockname()


class FTestWSGIHTTPServer(WSGIHTTPServer):

    def bind(self, addr):
        ip, port = addr
        super(FTestWSGIHTTPServer, self).bind(addr)
        if port == 0:
            ip, port = self.socket.getsockname()
            self.port = port


DOCTEST_EXAMPLE_FILENAME_RE = re.compile(r'<doctest (?P<name>.+)\[(?P<examplenum>\d+)\]>$')


def make_doctest_getlines_patch(test):
    orig_getlines = linecache.getlines
    def getlines(filename, module_globals=None):
        m = DOCTEST_EXAMPLE_FILENAME_RE.match(filename)
        if m and m.group('name') == test.name:
            example = test.examples[int(m.group('examplenum'))]
            source = example.source
            if isinstance(source, unicode):
                source = source.encode('ascii', 'backslashreplace')
            return source.splitlines(True)
        else:
            return orig_getlines(filename, module_globals)
    return getlines


def make_doctest_compile_patch(test):
    def compile_example(source, filename, mode, *args):
        match = DOCTEST_EXAMPLE_FILENAME_RE.match(filename)
        example_n = int(match.group('examplenum'))
        example = test.examples[example_n]
        line = example.lineno
        return compile('\n'*(line)+source, test.filename, mode, *args)
    return compile_example


def walk_the_traceback(traceback, test):
    """Try to find the traceback for this test."""
    try:
        test_filename = test.filename
    except AttributeError:
        return traceback

    while traceback.tb_next is not None:
        try:
            tb_filename = traceback.tb_frame.f_code.co_filename
            if tb_filename == test_filename:
                return traceback
        except AttributeError:
            pass
        traceback = traceback.tb_next

    return traceback


def acquire_testrunner_traceback(exc_info):
    ex_type, err, traceback = exc_info

    if isinstance(err, doctest.UnexpectedException):
        exc_info = err.exc_info
        traceback = exc_info[2]
    elif isinstance(err, doctest.DocTestFailure):
        try:
            last_line = max(0, len(err.example.source.splitlines())+err.example.lineno-1)
            throw_source = '\n'*(last_line) +\
                           'raise ValueError'\
                           '("Expected and actual output are different")'
            exec compile(throw_source, err.test.filename, 'single') in err.test.globs
        except:
            exc_info = sys.exc_info()
            traceback = exc_info[2]

    if isinstance(err, (doctest.DocTestFailure, doctest.UnexpectedException)):
        traceback = walk_the_traceback(traceback, err.test)

    return traceback


def walk_frames_to_find_output():
    from zope.testrunner.runner import TestResult
    ff = sys._getframe()
    while ff.f_back is not None:
        ff = ff.f_back
        if isinstance(ff.f_locals.get('self'), TestResult):
            testresult = ff.f_locals['self']
            return testresult.options.output



def report_testrunner_real_exception(exc_info):
    ex_type, err, traceback = exc_info
    if not isinstance(err, doctest.UnexpectedException):
        return # nothing to report
    try:
        import zope.testrunner.runner
    except ImportError:
        return # no testrunner
    exc_info = err.exc_info
    traceback = exc_info[2]
    traceback = walk_the_traceback(traceback, err.test)
    output = walk_frames_to_find_output()
    if output is not None:
        output.print_traceback(
            "Test failure:", (exc_info[0], exc_info[1], traceback))


def make_ipdb_testrunner_postmortem(test):
    import ipdb
    import zope.testrunner.interfaces
    def interactive_post_mortem(exc_info):
        report_testrunner_real_exception(exc_info)
        tb = acquire_testrunner_traceback(exc_info)
        ipdb.__main__.update_stdout()
        ipdb.__main__.BdbQuit_excepthook.excepthook_ori = sys.excepthook
        sys.excepthook = ipdb.__main__.BdbQuit_excepthook
        p = ipdb.__main__.Pdb(ipdb.__main__.def_colors)
        p.reset()
        p.botframe = tb.tb_frame
        p.interaction(tb.tb_frame, tb)
        raise zope.testrunner.interfaces.EndRun()
    return interactive_post_mortem


def make_pdb_testrunner_postmortem(test):
    import pdb
    import zope.testrunner.interfaces
    def interactive_post_mortem(exc_info):
        report_testrunner_real_exception(exc_info)
        tb = acquire_testrunner_traceback(exc_info)
        p = pdb.Pdb()
        p.reset()
        p.botframe = tb.tb_frame
        p.interaction(tb.tb_frame, tb)
        raise zope.testrunner.interfaces.EndRun()
    return interactive_post_mortem


def make_testrunner_postmortem_patch(test):
    try:
        return make_ipdb_testrunner_postmortem(test)
    except ImportError:
        pass

    try:
        return make_pdb_testrunner_postmortem(test)
    except ImportError:
        return None
    return None


class SeleniumLayer(ZCMLLayer):

    thread_count = 1
    ip = "127.0.0.1"
    port = 0
    serving = False

    config = None
    browsers = None
    downloads = None

    def __init__(self, *args, **kw):
        ZCMLLayer.__init__(self, *args, **kw)
        self.browsers = BrowserPool(self)
        self.downloads = Downloads(self)
        if BLACK_MAGIC:
            self.patches = []

    def setUp(self):
        if BLACK_MAGIC:
            self.patches.append(mock.ModulesSnapshot())
        global _Browser_UI_ext
        global _Element_UI_ext
        _Browser_UI_ext = ExtensionGroup()
        _Element_UI_ext = ExtensionGroup()
        ZCMLLayer.setUp(self)
        schooltool.testing.registry.setupSeleniumHelpers()
        self.setUpBrowserConfig()
        dispatcher = ThreadedTaskDispatcher()
        dispatcher.setThreadCount(self.thread_count)
        application = WSGIPublisherApplication(self.setup.db)
        self.server = FTestWSGIHTTPServer(application,
                                          'WSGI-HTTP',
                                          self.ip, self.port,
                                          task_dispatcher=dispatcher,
                                          verbose=True,
                                          hit_log=CommonAccessLogger())
        self.thread = threading.Thread(target=self.poll_server)
        self.thread.setDaemon(True)
        self.thread.start()
        print 'serving at http://%s:%s/' % (
            self.server.socket.getsockname())

    def setUpBrowserConfig(self):
        try:
            import schooltool.devtools.selenium_recipe
        except ImportError:
            return

        default = schooltool.devtools.selenium_recipe.default_browser_config
        config = default.copy()
        config.update(self.getScreenshotsConfig(default))
        config.update(self.getDownloadsConfig(default))
        self.config = config

    def getScreenshotsConfig(self, default):
        try:
            import schooltool.devtools.selenium_recipe
        except ImportError:
            return
        config = schooltool.devtools.selenium_recipe.BrowserConfig()

        if default.screenshots_dir:
            config.screenshots_dir = os.path.abspath(os.path.expanduser(default.screenshots_dir))

        if default.screenshots_url:
            config.screenshots_url = default.screenshots_url
        elif config.screenshots_dir:
            config.screenshots_url = 'file://%s/' % config.screenshots_dir

        config.overwrite_screenshots = default.overwrite_screenshots
        return config

    def getDownloadsConfig(self, default):
        try:
            import schooltool.devtools.selenium_recipe
        except ImportError:
            return
        config = schooltool.devtools.selenium_recipe.BrowserConfig()

        subdir = None

        if default.downloads_dir:
            config.downloads_dir = os.path.abspath(os.path.expanduser(default.downloads_dir))
            assert os.path.isdir(config.downloads_dir)
            i = 1
            while i < 10**9:
                subdir = 'testrun_%03d' % i
                if not os.path.exists(os.path.join(config.downloads_dir, subdir)):
                    break
                i = i + 1
            config.downloads_dir = os.path.join(config.downloads_dir, subdir)
            if not os.path.exists(config.downloads_dir):
                os.mkdir(config.downloads_dir)

        if default.downloads_url:
            url = schooltool.devtools.selenium_recipe.default_browser_config.downloads_url
            if subdir is not None:
                url = urllib.basejoin(url, subdir) + '/'
            config.downloads_url = url
        elif config.downloads_dir:
            config.downloads_url = 'file://%s/' % config.downloads_dir

        return config

    def tearDown(self):
        self.serving = False
        self.thread.join()
        self.browsers.reset()
        self.downloads.reset()
        self.config = None
        ZCMLLayer.tearDown(self)
        global _Browser_UI_ext
        global _Element_UI_ext
        _Browser_UI_ext = None
        _Element_UI_ext = None
        if BLACK_MAGIC:
            snapshot = self.patches.pop()
            snapshot.restore()
            assert not self.patches

    def poll_server(self):
        self.serving = True
        while self.serving:
            asyncore.poll(0.1)
        self.server.close()

    def testSetUp(self):
        if BLACK_MAGIC:
            self.patches.append(mock.ModulesSnapshot())
        self.browsers.reset()
        self.downloads.reset()
        self.setup.setUp()

        application = self.server.application
        factory = type(application.requestFactory)
        application.requestFactory = factory(self.setup.db)

    def testTearDown(self):
        self.setup.tearDown()
        self.browsers.reset()
        self.downloads.reset()
        if BLACK_MAGIC:
            snapshot = self.patches.pop()
            snapshot.restore()
            assert self.patches


_screenshot_test_short_name_map = {}


class SkippyDocTestRunner(doctest.DocTestRunner):
    """Mr. Skippy Runner.

    When REPORT_ONLY_FIRST_FAILURE is specified, Mr. Skippy skips
    the rest of each test when an error is encountered.  Needless to say,
    tests that have cleanup at the end should *never* have
    REPORT_ONLY_FIRST_FAILURE set.
    """

    def run(self, test, compileflags=None, out=None, clear_globs=True):
        result = None
        try:
            result = doctest.DocTestRunner.run(
                self, test, compileflags=compileflags,
                out=out, clear_globs=clear_globs)
        finally:
            for example in test.examples:
                if '__old_options' in example.__dict__:
                    example.options = example.__dict__['__old_options']
                    del example.__dict__['__old_options']
        return result

    def honour_only_first_failure_flag(self, test):
        if self.optionflags & doctest.REPORT_ONLY_FIRST_FAILURE:
            # For practical reasons, skip the rest of the test.
            # Selenium failures are slow enough as they are,
            # requiring to execute *all* examples in the test
            # is not an option considering the plausable implicit
            # waits on *each* failure.
            for example in test.examples:
                if '__old_options' not in example.__dict__:
                    example.__dict__['__old_options'] = example.options
                    example.options = example.options.copy()
                    example.options[doctest.SKIP] = True
                    del example.__dict__['__old_options']

    def report_unexpected_exception(self, out, test, example, exc_info):
        result = doctest.DocTestRunner.report_unexpected_exception(
            self, out, test, example, exc_info)
        self.output_screenshots(out, test, example, exc_info)
        self.honour_only_first_failure_flag(test)
        return result

    def report_failure(self, out, test, example, got):
        result = doctest.DocTestRunner.report_failure(
            self, out, test, example, got)
        self.output_screenshots(out, test, example, got)
        self.honour_only_first_failure_flag(test)
        return result

    def output_screenshots(self, out, test, example, got):
        layer = self._checker.layer
        if not layer.config.screenshots_dir:
            return

        screenshots = []
        browsers = {}
        if 'browsers' in test.globs:
            browser_pool = test.globs['browsers']
            browsers = browser_pool.browsers
        for browsername, browser in browsers.items():

            # Short test name map is used to get unique screenshot names,
            # when multiple tests with the same name are found (in multiple
            # modules)
            global _screenshot_test_short_name_map
            if test.name not in _screenshot_test_short_name_map:
                _screenshot_test_short_name_map[test.name] = {
                    test.filename: test.name}
                shortname = test.name
            else:
                shortnames = _screenshot_test_short_name_map[test.name]
                shortname = shortnames.get(test.filename)
                if shortname is None:
                    shortname = '%s-%d' % (test.name, len(shortnames) + 1)
                    shortnames[test.filename] = shortname

            basename = '%s-line_%d-%s' % (shortname, example.lineno, browsername)
            shot = browser.screenshot(basename, overwrite=layer.config.overwrite_screenshots)
            if shot:
                screenshots.append((shot.url or shot.path, shot.message))
            else:
                screenshots.append((shot.filename, shot.message))

        if not screenshots:
            return

        out('Screenshots:\n')
        out('\n'.join(['    %s%s' % (url, msg and (' - %s' % msg) or '')
                       for (url, msg) in screenshots]))


class RepeatyDebugRunner(doctest.DebugRunner, SkippyDocTestRunner):
    """Mr. Repeaty Debugger."""


class SeleniumOutputChecker(doctest.OutputChecker):

    def __init__(self, layer, *args, **kw):
        self.layer = layer

    def _unify_got(self, got):
        if (isinstance(got, unicode) or
            isinstance(got, str)):
            localhost = self.layer.browsers.localhost
            got = got.replace(localhost, 'http://localhost/')
        return got

    def check_output(self, want, got, optionflags):
        return doctest.OutputChecker.check_output(
            self, want, self._unify_got(got), optionflags)

    def output_difference(self, example, got, optionflags):
        return doctest.OutputChecker.output_difference(
            self, example, self._unify_got(got), optionflags)


class DiffTemplate(object):

    class ExampleStub(object):
        pass

    def __init__(self, checker, optionflags=None):
        self.checker = checker
        self.optionflags = optionflags

    def prettifyDiff(self, diff):
        output = walk_frames_to_find_output()
        if output is None:
            return diff
        old_stdout = sys.stdout
        result = StringIO()
        sys.stdout = result
        try:
            output.print_doctest_failure(diff)
        finally:
            sys.stdout = old_stdout
        return result.getvalue()

    def __mod__(self, params):
        file, line, name, source, want, got = params
        result = 'File "%s", line %s, in %s' % (file, line, name)
        result += '\n\n>>> '+'\n...'.join(source.splitlines())+'\n\n'
        example = self.ExampleStub()
        example.want = want
        diff = self.checker.output_difference(example, got, self.optionflags)
        result += self.prettifyDiff(diff)
        return result


def make_pretty_diff_template(*args, **kw):
    try:
        import zope.testrunner.formatter
    except ImportError:
        return None
    return DiffTemplate(*args, **kw)


class SeleniumDocFileCase(doctest.DocFileCase):

    def __init__(self, test,
                 optionflags=0,
                 setUp=None, tearDown=None,
                 checker=None):
        super(SeleniumDocFileCase, self).__init__(
            test, optionflags=optionflags,
            setUp=setUp, tearDown=tearDown, checker=checker)

    if BLACK_MAGIC:
        def patch(self, patches):
            self._layer.patches[-1].mock(
                dict(filter(lambda (a, m): m is not None, patches.items())))

    def run(self, *args, **kw):
        if BLACK_MAGIC:
            test = self._dt_test
            patches = {
                'doctest.compile':
                    make_doctest_compile_patch(test),
                }
            self.patch(patches)
        return doctest.DocFileCase.run(self, *args, **kw)

    def debug(self):
        if BLACK_MAGIC:
            test = self._dt_test
            patches = {
                'doctest.compile':
                    make_doctest_compile_patch(test),
                #'linecache.getlines':
                #    make_doctest_getlines_patch(test),
                'zope.testrunner.debug.post_mortem':
                    make_testrunner_postmortem_patch(test),
                'zope.testrunner.formatter.doctest_template':
                    make_pretty_diff_template(self._dt_checker,
                                              optionflags=self._dt_optionflags),
                }
            self.patch(patches)
        self.setUp()
        runner = RepeatyDebugRunner(optionflags=self._dt_optionflags,
                                    checker=self._dt_checker, verbose=False)
        runner.run(self._dt_test, clear_globs=False)
        self.tearDown()

    def runTest(self):
        test = self._dt_test
        old = sys.stdout
        new = StringIO()
        optionflags = self._dt_optionflags

        if not (optionflags & doctest.REPORTING_FLAGS):
            # The option flags don't include any reporting flags,
            # so add the default reporting flags
            optionflags |= doctest._unittest_reportflags

        runner = SkippyDocTestRunner(optionflags=optionflags,
                                     checker=self._dt_checker,
                                     verbose=False)

        try:
            runner.DIVIDER = "-"*70
            failures, tries = runner.run(
                test, out=new.write, clear_globs=False)
        finally:
            sys.stdout = old

        if failures:
            raise self.failureException(self.format_failure(new.getvalue()))


def SeleniumFileTest(path, module_relative=True, package=None,
                globs=None, parser=doctest.DocTestParser(),
                encoding=None, test_case_factory=SeleniumDocFileCase,
                checker=None,
                **options):
    globs = globs and globs.copy() or {}
    if package and not module_relative:
        raise ValueError("Package may only be specified for module-"
                         "relative paths.")

    doc, path = doctest._load_testfile(path, package, module_relative)

    if "__file__" not in globs:
        globs["__file__"] = path

    test_name = os.path.basename(path)

    if encoding is not None:
        doc = doc.decode(encoding)

    test = parser.get_doctest(doc, globs, test_name, path, 0)
    return test_case_factory(test, checker=checker, **options)


def SeleniumDocFileSuite(layer, *paths, **kw):
    if kw.get('optionflags') is None:
        kw['optionflags'] = (doctest.ELLIPSIS |
                             doctest.REPORT_NDIFF |
                             doctest.NORMALIZE_WHITESPACE |
                             doctest.REPORT_ONLY_FIRST_FAILURE)
    suite = unittest.TestSuite()
    for path in paths:
        suite.addTest(SeleniumFileTest(
                path, checker=SeleniumOutputChecker(layer), **kw))
    suite.layer = layer
    return suite


def collect_ftests(package=None, level=None, layer=None, filenames=None,
                   optionflags=None, test_case_factory=SeleniumDocFileCase,
                   extra_globs=None):
    package = doctest._normalize_module(package)
    extra_globs = extra_globs or {}
    extra_globs.update({'browsers': layer.browsers,
                        'downloads': layer.downloads})
    def make_suite(filename, package=None):
        if BLACK_MAGIC:
            def layer_setting_factory(*args, **kw):
                test_case = test_case_factory(*args, **kw)
                test_case._layer = layer
                return test_case
            suite_test_case_factory = layer_setting_factory
        else:
            suite_test_case_factory = test_case_factory
        suite = SeleniumDocFileSuite(layer, filename,
                                     package=package,
                                     optionflags=optionflags,
                                     globs=extra_globs,
                                     encoding='UTF-8',
                                     test_case_factory=suite_test_case_factory)
        return suite
    assert isinstance(layer, SeleniumLayer)
    suite = collect_txt_ftests(package=package, level=level,
                               layer=layer, filenames=filenames,
                               suite_factory=make_suite)
    if not selenium_enabled:
        # XXX: should log that tests are skipped somewhere better
        print >> sys.stderr, 'Selenium not configured, skipping', package.__name__
        return unittest.TestSuite()
    return suite


class PreferNoLanguage(object):

    def __init__(self, context):
        pass

    def getPreferredLanguages(self):
        return ()



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Emergency Day Views
"""
import datetime

from zope.event import notify
from zope.i18n import translate
from zope.publisher.browser import BrowserView
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.traversing.browser.absoluteurl import absoluteURL

from schooltool.common import SchoolToolMessage as _
from schooltool.app.cal import CalendarEvent
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.interfaces import ISchoolToolCalendar
from schooltool.calendar.utils import parse_date
from schooltool.term.interfaces import ITerm
from schooltool.term.term import EmergencyDayEvent


class EmergencyDayView(BrowserView):
    """On emergencies such as extreme temperetures, blizzards, etc,
    school is cancelled, and a different day gets added to the term
    instead.

    This view lets the administrator choose the cancelled date and
    presents the user with a choice of non-schooldays within a term
    and the days immediately after the term.
    """

    __used_for__ = ITerm

    template = None
    date_template = ViewPageTemplateFile('templates/emergency_select.pt')
    replacement_template = ViewPageTemplateFile('templates/emergency2.pt')

    error = None
    date = None
    replacement = None

    def replacements(self):
        """Return all non-schooldays in term plus 3 days after the term."""
        result = []
        for date in self.context:
            if date > self.date:
                if not self.context.isSchoolday(date):
                    result.append(date)
        last = self.context.last
        day = datetime.timedelta(1)
        result.append(last + day)
        result.append(last + 2 * day)
        result.append(last + 3 * day)
        return result

    def update(self):
        self.template = self.date_template
        if 'CANCEL' in self.request:
            self.request.response.redirect(
                absoluteURL(self.context, self.request))
            return
        if 'date' in self.request:
            try:
                self.date = parse_date(self.request['date'])
            except ValueError:
                self.error = _("The date you entered is invalid."
                               "  Please use the YYYY-MM-DD format.")
                return
            if not self.date in self.context:
                self.error = _("The date you entered does not belong to"
                               " this term.")
                return
            if not self.context.isSchoolday(self.date):
                self.error = _("The date you entered is not a schoolday.")
                return

            self.template = self.replacement_template
        if 'replacement' in self.request:
            try:
                self.replacement = parse_date(self.request['replacement'])
            except ValueError:
                self.error = _("The replacement date you entered is invalid.")
                self.template = self.replacement_template
                return

        if self.date and self.replacement:
            if self.context.last < self.replacement:
                # XXX: I wonder if all places that are dependent on
                #      term start/end dates are updated properly
                self.context.last = self.replacement

            # XXX: assert?!
            assert not self.context.isSchoolday(self.replacement)
            assert self.context.isSchoolday(self.date)
            self.context.add(self.replacement)

            notify(EmergencyDayEvent(self.date, self.replacement))

            # XXX: Following code should be move to the event subscriber, but!
            #      It wants to store translated messages, and in current default
            #      case the event description will be stored in the language
            #      from somebodys browser settings.  Now that's naughty!
            #      If only we had 'apllication language' separate form 'user
            #      presentation language' this problem would go away.
            #
            # Post calendar events to schoolwide calendar
            calendar = ISchoolToolCalendar(ISchoolToolApplication(None))
            dtstart = datetime.datetime.combine(self.date, datetime.time())
            msg = _('School cancelled due to emergency.'
                    ' Replacement day $replacement.',
                    mapping={'replacement': str(self.replacement)})
            msg = translate(msg, context=self.request)
            calendar.addEvent(
                CalendarEvent(dtstart, datetime.timedelta(),
                              msg, allday=True))

            dtstart = datetime.datetime.combine(self.replacement,
                                                datetime.time())
            msg = _('Replacement day for emergency day $emergency.',
                    mapping={'emergency': str(self.date)})
            msg = translate(msg, context=self.request)
            calendar.addEvent(
                CalendarEvent(dtstart, datetime.timedelta(),
                              msg, allday=True))

            self.request.response.redirect(
                absoluteURL(self.context, self.request))

    def __call__(self):
        self.update()
        return self.template()





#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Flourish error views.
"""
from zope.app.exception.browser.unauthorized import Unauthorized

from schooltool.skin.error import ErrorView
from schooltool.skin.flourish.page import NoSidebarPage


class NotFound(object):

    def __call__(self, *args, **kw):
        self.request.response.setStatus(404)
        return super(NotFound, self).__call__(*args, **kw)


class ErrorPage(NoSidebarPage, ErrorView):
    container_class = 'container extra-wide-container'

    def __call__(self, *args, **kw):
        self.request.response.setStatus(500)
        return NoSidebarPage.__call__(self, *args, **kw)


class UnauthorizedPage(NoSidebarPage, Unauthorized):

    def render(self, *args, **kw):
        return Unauthorized.__call__(self)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005-2010 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool application
"""
import calendar
import os
import pytz

from persistent import Persistent
from persistent.dict import PersistentDict

from zope.location.location import LocationProxy
from zope.component import adapter, adapts
from zope.i18n import translate
from zope.interface import implementer, implements, implementsOnly

from zope.annotation.interfaces import IAttributeAnnotatable, IAnnotations
from zope.app.applicationcontrol.interfaces import IApplicationControl
from zope.app.applicationcontrol.applicationcontrol import applicationController
from zope.component.hooks import getSite
from zope.site import SiteManagerContainer
from zope.container import sample
from zope.container.contained import NameChooser
from zope.container.interfaces import INameChooser
from zope.traversing.interfaces import IContainmentRoot

from schooltool.app.interfaces import IPluginInit, IPluginStartUp
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.interfaces import IApplicationPreferences, IApplicationTabs
from schooltool.app import relationships
from schooltool.app.interfaces import IAsset
from schooltool.relationship.relationship import RelationshipProperty
from schooltool.common import getRequestFromInteraction
from schooltool.common import SchoolToolMessage as _


class SchoolToolApplication(Persistent, sample.SampleContainer,
                            SiteManagerContainer):
    """The main application object.

    This object can be added as a regular content object to a folder,
    or it can be used as the application root object.
    """

    implements(ISchoolToolApplication, IAttributeAnnotatable, IContainmentRoot)

    def __init__(self):
        super(SchoolToolApplication, self).__init__()

    def _newContainerData(self):
        return PersistentDict()

    def _title(self):
        """This is required for the site calendar views to work.

        Calendar views expect that their underlying objects (calendar
        parents) have an attribute named `title`.
        """
        return IApplicationPreferences(self).title

    title = property(_title)


@adapter(None)
@implementer(ISchoolToolApplication)
def getSchoolToolApplication(ignore=None):
    """Return the nearest ISchoolToolApplication.

    This function is also registered as an adapter, so you
    can use it like this:

        app = ISchoolToolApplication(None)

    and stub it in unit tests.
    """
    candidate = getSite()
    if ISchoolToolApplication.providedBy(candidate):
        return candidate
    return None


class SimpleNameChooser(NameChooser):
    """A name chooser that uses object titles as names.

    SimpleNameChooser is an adapter for containers

        >>> class ContainerStub(dict):
        ...     __name__ = 'resources'
        >>> container = ContainerStub()
        >>> chooser = SimpleNameChooser(container)

    It expects objects to have a `title` attribute, so only register it
    as an adapter for containers that limit their contents to objects
    with such attribute.

    `chooseName` uses the title of the object to be added, converts it to
    lowercase, strips punctuation, and replaces spaces with hyphens:

        >>> from schooltool.person.person import Person
        >>> obj = Person(title='Mr. Smith')
        >>> chooser.chooseName('', obj)
        u'mr-smith'

    If the name is already taken, SimpleNameChooser adds a number at the end

        >>> container['mr-smith'] = 42
        >>> chooser.chooseName('', obj)
        u'mr-smith-2'

    If you provide a suggested name, it uses that one.

        >>> chooser.chooseName('suggested-name', obj)
        'suggested-name'

    Even if it has "illegal" characters:

        >>> chooser.chooseName('suggested name', obj)
        'suggested name'

    But it will add a number at the end:

        >>> chooser.chooseName('mr-smith', obj)
        u'mr-smith-2'

    Bad names cause errors

        >>> chooser.chooseName('@notallowed', obj)
        Traceback (most recent call last):
          ...
        ValueError: Names cannot begin with '+' or '@' or contain '/'

    If the name generated from the title gets shortened too much, we
    generate a name from the name of the context container instead:

        >>> obj.title = "foo-bar-baz-boo"
        >>> chooser.chooseName('', obj)
        'resource'

    So objects with semi empty titles get ids for them too:

        >>> obj.title = '---'
        >>> chooser.chooseName('', obj)
        'resource'

    """

    implements(INameChooser)

    def chooseName(self, name, obj):
        """See INameChooser."""
        if not name:
            name = u''.join([c for c in obj.title.lower()
                             if c.isalnum() or c == ' ']).replace(' ', '-')
            if len(name) + 2 < len(obj.title) or name == '':
                name = self.context.__name__[:-1]
        n = name
        i = 1
        while n in self.context:
            i += 1
            n = name + u'-' + unicode(i)
        # Make sure the name is valid
        self.checkName(n, obj)
        return n


class ApplicationPreferences(Persistent):
    """Object for storing any application-wide preferences we have.

    See schooltool.app.interfaces.ApplicationPreferences.
    """

    implements(IApplicationPreferences)

    __name__ = 'preferences'

    def getTitle(self):
        title = self.__dict__.get('title', None)
        if title:
            return title
        request = getRequestFromInteraction()
        return translate(_('Your School'), context=request)

    def setTitle(self, value):
        self.__dict__['title'] = value

    title = property(getTitle, setTitle)

    timezone = 'UTC'

    # XXX: initialize from locale
    dateformat = '%Y-%m-%d'
    timeformat = '%H:%M'
    weekstart = calendar.MONDAY

    frontPageCalendar = True

    logo = None

    name_sorting = 'last_name'

    def initializeTimezone(self):
        self.timezone = self._get_localzone()

    def _get_localzone(self):
        """Tries to find the local timezone configuration.

        Taken from https://github.com/regebro/tzlocal/
        only the part needed for Ubuntu.
        """
        tzpath = '/etc/timezone'
        if os.path.exists(tzpath):
            with open(tzpath, 'rb') as tzfile:
                data = tzfile.read()
                if data[:5] != 'TZif2':
                    etctz = data.strip().decode()
                    # Get rid of host definitions and comments:
                    if ' ' in etctz:
                        etctz, dummy = etctz.split(' ', 1)
                    if '#' in etctz:
                        etctz, dummy = etctz.split('#', 1)
                    if 'Etc/UTC' == etctz:
                        etctz = 'UTC'
                    try:
                        pytz.timezone(etctz)
                        return etctz
                    except pytz.UnknownTimeZoneError:
                        pass

        return 'UTC'


class ApplicationTabs(PersistentDict):
    """Object for storing application tab preferences.

    See schooltool.app.interfaces.ApplicationTabs.
    """

    implements(IApplicationTabs)

    default = 'calendar'


class Asset(object):
    """A mixin for objects that may act as assets."""

    implements(IAsset)

    leaders = RelationshipProperty(relationships.URILeadership,
                                   relationships.URIAsset,
                                   relationships.URILeader)


def getApplicationPreferences(app):
    """Adapt a SchoolToolApplication to IApplicationPreferences."""

    annotations = IAnnotations(app)
    key = 'schooltool.app.ApplicationPreferences'
    try:
        return annotations[key]
    except KeyError:
        annotations[key] = ApplicationPreferences()
        annotations[key].__parent__ = app
        return annotations[key]


def getApplicationTabs(app):
    """Adapt a SchoolToolApplication to IApplicationTabs."""

    annotations = IAnnotations(app)
    key = 'schooltool.app.ApplicationTabs'
    try:
        return annotations[key]
    except KeyError:
        annotations[key] = ApplicationTabs()
        annotations[key].__parent__ = app
        return annotations[key]


class SchoolToolInitializationUtility(object):

    def initializeApplication(self, app):
        """Perform school specific initialization.

        By default schooltool does not do any specific initialization.
        """


class ActionBase(object):
    adapts(ISchoolToolApplication)
    implements(IPluginInit)

    after = ()
    before = ()

    def __init__(self, app):
        self.app = app

    def __call__(self):
        raise NotImplementedError("This method should be overriden by"
                                  " inheriting classes")


class InitBase(ActionBase):
    implementsOnly(IPluginInit)


class StartUpBase(ActionBase):
    implementsOnly(IPluginStartUp)


@adapter(ISchoolToolApplication)
@implementer(IApplicationControl)
def getApplicationControl(app=None):
    return LocationProxy(applicationController, app, 'control')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2007 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
BasicPerson specific security code.
"""

from schooltool.basicperson.interfaces import IBasicPerson
from schooltool.common import SchoolToolMessage as _
from schooltool.person.interfaces import IPerson
from schooltool.securitypolicy.crowds import Crowd


class PersonAdvisorsCrowd(Crowd):
    """Crowd of advisors of a person."""

    title = _(u'Advisors')
    description = _(u'Advisors of a person.')

    def contains(self, principal):
        user = IPerson(principal, None)
        person = self.context
        if not IBasicPerson.providedBy(person):
            return False
        return user in person.advisors




#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Base classes for report reference and request adapters

"""

import urllib

try:
    from kombu.utils import symbol_by_name
except ImportError:
    from celery.utils import get_symbol_by_name as symbol_by_name

import zope.i18n.locales
import zope.component
import zope.interface
import zope.publisher.base
import zope.file.file
from persistent.dict import PersistentDict
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.component import getUtility, queryUtility, getGlobalSiteManager
from zope.interface import implements
from zope.intid.interfaces import IIntIds
from zope.publisher.browser import BrowserView
from zope.publisher.browser import BrowserRequest
from zope.publisher.http import HTTPResponse
from zope.traversing.browser.absoluteurl import absoluteURL
from z3c.rml import rml2pdf

import schooltool.common
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.course.interfaces import ISectionContainer
from schooltool.group.interfaces import IGroupContainer
from schooltool.report.interfaces import IReportLinkViewletManager
from schooltool.report.interfaces import IReportLinkViewlet
from schooltool.report.interfaces import IRegisteredReportsUtility
from schooltool.report.interfaces import IReportLinksURL
from schooltool.report.interfaces import IReportTask
from schooltool.report.interfaces import IReportDetails
from schooltool.report.interfaces import IReportMessage
from schooltool.report.interfaces import IReportProgressMessage
from schooltool.report.interfaces import IRemoteReportLayer
from schooltool.report.interfaces import IReportFile
from schooltool.schoolyear.interfaces import ISchoolYear
from schooltool.skin import flourish
from schooltool.task.tasks import RemoteTask
from schooltool.task.tasks import query_messages
from schooltool.task.tasks import Message
from schooltool.task.progress import ProgressMessage
from schooltool.task.interfaces import ITaskScheduledNotification
from schooltool.task.tasks import TaskCompletedNotification
from schooltool.task.tasks import TaskScheduledNotification
from schooltool.task.tasks import TaskFailedMessage
from schooltool.term.interfaces import IDateManager

from schooltool.common import SchoolToolMessage as _


class ReportFile(zope.file.file.File):
    implements(IReportFile)


class ReportLinkViewletManager(flourish.viewlet.ViewletManager):
    implements(IReportLinkViewletManager)


class IFlourishReportLinkViewletManager(flourish.interfaces.IViewletManager,
                                        IReportLinkViewletManager):
    pass


class FlourishReportLinkViewletManager(flourish.viewlet.ViewletManager):
    template = ViewPageTemplateFile('templates/f_report_link_manager.pt')

    @property
    def table(self):
        result = {}
        for viewlet in self.viewlets:
            group = result.setdefault(viewlet.file_type, {
                'file_type': viewlet.file_type.upper(),
                'rows': [],
                })
            group['rows'].append({
                'title': viewlet.title,
                'url': viewlet.link,
                'link_id': viewlet.link.replace('.', '_'),
                'form_id': viewlet.link.replace('.', '_') + '_form',
                'description': viewlet.description,
                })
        return [group for key, group in sorted(result.items())]


class ReportLinkViewlet(object):

    implements(IReportLinkViewlet)

    template=ViewPageTemplateFile('templates/report_link.pt')
    group=u''
    title=u''
    link=u'' # an optional relative link - subclasses can override report_link property in some cases

    @property
    def report_link(self):
        return '%s/%s' % (absoluteURL(self.context, self.request), self.link)

    def render(self, *args, **kw):
        return self.template()


class RegisteredReportsUtility(object):
    implements(IRegisteredReportsUtility)

    def __init__(self):
        self.reports_by_group = {}

    def registerReport(self, group, title, description, file_type, name, layer):
        # make a non-translatable group key
        group_key = unicode(group)

        if group_key not in self.reports_by_group:
            self.reports_by_group[group_key] = []
        self.reports_by_group[group_key].append({
            'group': group, # remember the translatable group title
            'title': title,
            'description': description,
            'file_type': file_type,
            'name': name,
            'layer': layer,
            })


def getReportRegistrationUtility():
    """Helper - returns report registration utility and registers a new one
    if missing."""
    utility = queryUtility(IRegisteredReportsUtility)
    if not utility:
        utility = RegisteredReportsUtility()
        getGlobalSiteManager().registerUtility(utility,
            IRegisteredReportsUtility)
    return utility


class ReportLinksURL(BrowserView):
    implements(IReportLinksURL)

    def actualContext(self):
        return self.context

    def __unicode__(self):
        return urllib.unquote(self.__str__()).decode('utf-8')

    def __str__(self):
        return absoluteURL(self.actualContext(), self.request)

    def __call__(self):
        return self.__str__()


class StudentReportLinksURL(ReportLinksURL):

    def actualContext(self):
        return ISchoolToolApplication(None)['persons']


class GroupReportLinksURL(ReportLinksURL):

    def actualContext(self):
        current_term = getUtility(IDateManager).current_term
        if current_term is None:
            return ISchoolToolApplication(None)
        return IGroupContainer(ISchoolYear(current_term))


class SchoolYearReportLinksURL(ReportLinksURL):

    def actualContext(self):
        current_term = getUtility(IDateManager).current_term
        if current_term is None:
            return ISchoolToolApplication(None)
        return ISchoolYear(current_term)


class TermReportLinksURL(ReportLinksURL):

    def actualContext(self):
        current_term = getUtility(IDateManager).current_term
        if current_term is None:
            return ISchoolToolApplication(None)
        return current_term


class SectionReportLinksURL(ReportLinksURL):

    def actualContext(self):
        current_term = getUtility(IDateManager).current_term
        if current_term is None:
            return ISchoolToolApplication(None)
        return ISectionContainer(current_term)


class FlourishSchoolReportLinksURL(ReportLinksURL):

    def __str__(self):
        app = ISchoolToolApplication(None)
        return absoluteURL(app, self.request) + '/manage'


class FlourishPersonReportLinksURL(ReportLinksURL):

    def __str__(self):
        app = ISchoolToolApplication(None)
        return absoluteURL(app, self.request) + '/persons'


class FlourishGroupReportLinksURL(GroupReportLinksURL):

    def __str__(self):
        app = ISchoolToolApplication(None)
        return absoluteURL(app, self.request) + '/groups'


class FlourishSchoolYearReportLinksURL(SchoolYearReportLinksURL):

    def __str__(self):
        app = ISchoolToolApplication(None)
        return absoluteURL(app, self.request) + '/schoolyears'


class FlourishTermReportLinksURL(TermReportLinksURL):

    def __str__(self):
        app = ISchoolToolApplication(None)
        return absoluteURL(app, self.request) + '/terms'


class FlourishSectionReportLinksURL(SectionReportLinksURL):

    def __str__(self):
        app = ISchoolToolApplication(None)
        return absoluteURL(app, self.request) + '/sections'


class FlourishResourceReportLinksURL(ReportLinksURL):

    def __str__(self):
        app = ISchoolToolApplication(None)
        return absoluteURL(app, self.request) + '/resources'


class RemoteReportResponse(HTTPResponse):
    pass


class ReportResourceURL(object):
    zope.component.adapts(IRemoteReportLayer)
    zope.interface.implements(schooltool.common.IResourceURIGetter)

    def __init__(self, request):
        self.request = request

    def __call__(self, library_name, resource_name):
        if not resource_name:
            return None
        fake_request = BrowserRequest('', {})
        zope.interface.directlyProvides(
            fake_request, flourish.interfaces.IFlourishLayer)
        if library_name is not None:
            library = zope.component.queryAdapter(
                fake_request, name=library_name)
            resource = library.get(resource_name)
        else:
            resource = zope.component.queryAdapter(
                fake_request, name=resource_name)
        if resource is None:
            return None
        result = schooltool.common.data_uri(
            resource.context.data,
            mime=resource.context.content_type)
        return result


class RemoteRequestParams(PersistentDict):
    locale_id = None
    cookies = None


class RemoteReportRequest(object):
    implements(IRemoteReportLayer)

    response = None
    debug = None
    annotations = None
    locale = None
    interaction = None
    form = None
    URL = ''
    task = None
    cookies = None

    def __init__(self, task=None):
        self.task = task
        self.response = RemoteReportResponse()
        self.debug = zope.publisher.base.DebugFlags()
        self.form = {}
        self.annotations = {}
        self.cookies = {}

    def initParams(self, params):
        self.form.update(params)
        locale_id = getattr(params, 'locale_id', None)
        if locale_id is not None:
            self.locale = zope.i18n.locales.Locale(locale_id)
        cookies = getattr(params, 'cookies', None)
        if self.cookies is not None and cookies is not None:
            self.cookies.update(cookies)

    def clone(self):
        clone = self.__class__(task=self.task)
        clone.initParams(self.form)
        if self.locale is not None:
            clone.locale = zope.i18n.locales.Locale(self.locale.id)
        clone.cookies = dict(self.cookies)
        return clone

    @property
    def task_id(self):
        if self.task is None:
            return None
        return self.task.task_id

    def get(self, value, default=None):
        return self.form.get(value, default)

    def __getitem__(self, name):
        return self.form[name]

    def __contains__(self, name):
        return name in self.form

    def getVirtualHostRoot(self):
        return None

    def getApplicationURL(self, depth=0, path_only=False):
        return ''


class NoReportException(Exception):
    pass


class AbstractReportTask(RemoteTask):
    implements(IReportTask)

    routing_key = "zodb.report"
    default_mimetype = None
    default_filename = 'report'
    report = None

    abstract = True

    view_name = None
    factory_name = None

    context_intid = None

    def __init__(self, report_builder, context, remote_request=None):
        RemoteTask.__init__(self)
        if isinstance(report_builder, str):
            self.view_name = report_builder
        else:
            self.factory = report_builder
        self.context = context
        if remote_request is None:
            self.request_params = RemoteRequestParams()
        else:
            self.request_params = remote_request

    def default_request_param(self, name, other):
        if (name not in self.request_params and
            name in other):
            self.request_params[name] = other[name]

    def getCookies(self, request):
        cookies = {}
        for name in request.cookies:
            # Skip Zope's session cookies
            if 'zope3_cs' in name.lower():
                continue
            cookies[name] = request.cookies[name]
        return cookies

    def update(self, request):
        super(AbstractReportTask, self).update(request)
        if (self.request_params.locale_id is None and
            request.locale is not None):
            self.request_params.locale_id = request.locale.id
        self.default_request_param('HTTP_ACCEPT_LANGUAGE', request)
        self.request_params.cookies = dict(self.getCookies(request))

    def beginRequest(self, params=None):
        # XXX: move begin/end requests to runTransaction level
        if params is None:
            params = self.request_params
        request = RemoteReportRequest(task=self)
        if params is not None:
            request.initParams(params)
        # XXX:
        from schooltool.app.security import Principal
        from zope.security.checker import ProxyFactory
        principal = Principal(self.creator.__name__, 'XXX:title',
                              person=ProxyFactory(self.creator))
        for group in self.creator.groups:
            principal.groups.append("sb.group." + group.__name__)
        request.principal = principal
        zope.security.management.endInteraction()
        zope.security.management.newInteraction(request)

    def endRequest(self, request=None):
        zope.security.management.endInteraction()

    def getCurrentRequest(self):
        policy = zope.security.management.queryInteraction()
        if policy is None:
            return None
        requests = getattr(policy, 'participations', ())
        if not requests:
            return None
        return requests[0]

    def execute(self, celery_task, *args, **kwargs):
        self.beginRequest()
        renderer = self.getRenderer()
        if renderer is None:
            return # skip the report
        report_file = self.renderToFile(renderer, *args, **kwargs)
        self.report = report_file
        self.endRequest()

    def complete(self, request, result):
        self.beginRequest()
        res = super(AbstractReportTask, self).complete(request, result)
        self.endRequest()
        return res

    def fail(self, request, result, traceback):
        self.beginRequest()
        res = super(AbstractReportTask, self).fail(request, result, traceback)
        self.endRequest()
        return res

    def getRenderer(self, request=None):
        context = self.context
        if context is None:
            return None
        if request is None:
            request = self.getCurrentRequest()
        renderer = None
        factory = self.factory
        if factory is not None:
            renderer = factory(context, request)
        else:
            renderer = zope.component.queryMultiAdapter(
                (context, request), name=self.view_name)
        return renderer

    def renderReport(self, renderer, stream, *args, **kw):
        data = renderer()
        stream.write(data)

    def updateReport(self, renderer, report):
        if not report.mimeType:
            report.mimeType = getattr(renderer, 'mimetype', None)
        if not report.mimeType:
            report.mimeType = self.default_mimetype
        if not report.__name__ or not report.__name__.strip():
            report.__name__ = getattr(renderer, 'filename', None)
        if not report.__name__ or not report.__name__.strip():
            report.__name__ = self.default_filename

    def renderToFile(self, renderer, *args, **kw):
        report = ReportFile()
        stream = report.open('w')
        try:
            self.renderReport(renderer, stream, *args, **kw)
        except NoReportException:
            return None
        finally:
            stream.close()
        self.updateReport(renderer, report)
        return report

    @property
    def factory(self):
        result = symbol_by_name(self.factory_name)
        return result

    @factory.setter
    def factory(self, value):
        try:
            self.factory_name = '%s:%s' % (value.__module__, value.__name__)
        except AttributeError, e:
            raise e

    @property
    def context(self):
        int_ids = getUtility(IIntIds)
        obj = int_ids.queryObject(self.context_intid)
        return obj

    @context.setter
    def context(self, value):
        int_ids = getUtility(IIntIds)
        intid = int_ids.getId(value)
        self.context_intid = intid


class ReportTask(AbstractReportTask):

    default_filename = 'report.pdf'
    default_mimetype = 'application/pdf'

    def renderReport(self, renderer, stream, *args, **kw):
        renderer.update()
        rml = renderer.render()
        filename = renderer.filename
        pdf = rml2pdf.parseString(rml, filename=filename or None)
        stream.write(pdf.getvalue())


class OldReportTask(ReportTask):

    filename = None

    def renderReport(self, renderer, stream, *args, **kw):
        filename, data = renderer.renderToFile()
        if data is None:
            raise NoReportException()
        self.filename = filename
        stream.write(data)

    def updateReport(self, renderer, report):
        if not report.__name__ or not report.__name__.strip():
            report.__name__ = self.filename
        if not report.__name__ or not report.__name__.strip():
            report.__name__ = self.default_filename
        return ReportTask.updateReport(renderer, report)


class ArchiveReportTask(AbstractReportTask):

    default_filename = 'report.zip'
    default_mimetype = 'application/zip'

    def renderReport(self, renderer, stream, *args, **kw):
        renderer.update()
        written = renderer.render(stream, *args, **kw)
        if not written:
            raise NoReportException()


class ReportDetails(object):
    implements(IReportDetails)

    report = None
    requested_on = None
    group = _("Reports")
    filename = None

    def __init__(self, requested_on=None, filename=None):
        self.requested_on = requested_on
        self.filename = filename


class ReportProgressMessage(ProgressMessage, ReportDetails):
    implements(IReportProgressMessage)

    def __init__(self, title=None,
                 requested_on=None, filename=None):
        ProgressMessage.__init__(self, title=title)
        ReportDetails.__init__(
            self, requested_on=requested_on,
            filename=(filename or 'report.pdf'))


class ReportMessage(Message, ReportDetails):
    implements(IReportMessage)

    def __init__(self, title=None,
                 requested_on=None, filename=None):
        Message.__init__(self, title=title)
        ReportDetails.__init__(
            self, requested_on=requested_on,
            filename=(filename or 'report.pdf'))


class OnReportScheduled(TaskScheduledNotification):

    def __init__(self, task, http_request):
        super(TaskScheduledNotification, self).__init__(
            task, http_request)
        self.request = RemoteReportRequest()
        self.request.initParams(task.request_params)

    def send(self):
        renderer = self.task.getRenderer(request=self.request)
        subscribers = zope.component.getAdapters(
            (self.task, self.request, renderer),
            ITaskScheduledNotification)
        for name, subscriber in subscribers:
            subscriber(name=name)


class ReportFailedMessage(ReportMessage, TaskFailedMessage):

    def __init__(self, task, requested_on=None, filename=None):
        ReportMessage.__init__(self, requested_on=requested_on, filename=filename)
        TaskFailedMessage.__init__(self, task)


class OnPDFReportScheduled(TaskScheduledNotification):

    view = None
    message_factory = ReportProgressMessage

    def __init__(self, task, request, view):
        super(OnPDFReportScheduled, self).__init__(task, request)
        self.view = view

    def makeReportTitle(self):
        title = getattr(self.view, 'message_title', None)
        if not title:
            title = getattr(self.view, 'filename', None)
        if not title:
            title = _(u'report')
        return title

    def send(self):
        view = self.view
        view.render_invariant = True
        task = self.task
        title = self.makeReportTitle()
        msg = self.message_factory(
            title=title,
            requested_on=task.scheduled,
            filename=view.filename,
            )
        msg.send(sender=task, recipients=[self.task.creator])


class OnReportArchiveScheduled(OnPDFReportScheduled):
    pass


class GeneratedReportMessage(ReportMessage):

    def __init__(self, *args, **kw):
        if not kw.get('title'):
            kw['title'] = kw.get('filename', _('report'))
        report = kw.pop('report', None)
        super(GeneratedReportMessage, self).__init__(*args, **kw)
        self.report = report
        if self.report is not None:
            self.report.__parent__ = self


class OnReportGenerated(TaskCompletedNotification):

    message_factory = GeneratedReportMessage

    @property
    def messages(self):
        messages = query_messages(self.task)
        return messages

    def send(self):
        messages = self.messages
        for message in messages:
            generated_msg = self.message_factory(
                title=getattr(message, 'title', None),
                requested_on=message.requested_on,
                filename=message.filename,
                report = self.task.report)
            generated_msg.updated_on = self.task.utcnow
            generated_msg.replace(
                message,
                sender=self.task.creator,
                recipients=message.recipients)
            # XXX: set message expiration date
            # message.expires_on =
